import os


def wf_clickhouse_host() -> str:
    """The host of the clickhouse server."""
    return os.environ.get("WF_CLICKHOUSE_HOST", "localhost")


def wf_clickhouse_port() -> int:
    """The port of the clickhouse server."""
    return int(os.environ.get("WF_CLICKHOUSE_PORT", 8123))


def wf_clickhouse_user() -> str:
    """The user of the clickhouse server."""
    return os.environ.get("WF_CLICKHOUSE_USER", "default")


def wf_clickhouse_pass() -> str:
    """The password of the clickhouse server."""
    return os.environ.get("WF_CLICKHOUSE_PASS", "")


def wf_clickhouse_database() -> str:
    """The name of the clickhouse database."""
    return os.environ.get("WF_CLICKHOUSE_DATABASE", "default")



"""The top-level functions and classes for working with Weave."""

import sys

# We track what modules were loaded before importing weave, so we can ensure
# that someone doesn't introduce auto-importing loading weave.legacy.weave.ops or weave.legacy.weave.panels
# (because they are slow to import and have more dependencies, and they are part of
# the engine and UI layers which should be kept separate from the core layer).
pre_init_modules = set(sys.modules.keys())

from weave.legacy.weave import context_state as _context_state

_loading_builtins_token = _context_state.set_loading_built_ins()

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import storage
from weave.legacy.weave.api import *
from weave.legacy.weave.errors import *
from weave.legacy.weave import mappers_python_def
from weave.legacy.weave import wandb_api as _wandb_api
from weave.legacy.weave import context as _context

from weave import version

_wandb_api.init()

# Ensure there is a client available for eager mode


from weave.trace.api import *
_context_state.clear_loading_built_ins(_loading_builtins_token)

__version__ = version.VERSION

from weave.flow.obj import Object
from weave.flow.dataset import Dataset
from weave.flow.model import Model
from weave.flow.eval import Evaluation, Scorer
from weave.flow.agent import Agent, AgentState
from weave.trace.util import ThreadPoolExecutor, Thread

# See the comment above pre_init_modules above. This is check to ensure we don't accidentally
# introduce loading weave.legacy.weave.ops or weave.legacy.weave.panels when importing weave.
newly_added_modules = set(sys.modules.keys()) - pre_init_modules
ops_modules = []
panels_modules = []
for module_name in newly_added_modules:
    if module_name.startswith("weave.legacy.weave.ops"):
        ops_modules.append(module_name)
    if module_name.startswith("weave.legacy.weave.panels"):
        panels_modules.append(module_name)
if ops_modules or panels_modules:
    all_invalid_modules = ops_modules + panels_modules
    invalid_submodules = set([".".join(m.split(".")[:2]) for m in all_invalid_modules])
    raise errors.WeaveInternalError(
        "importing weave should not import weave.legacy.weave.ops or weave.legacy.weave.panels, but the following modules were imported: "
        + ", ".join(invalid_submodules)
    )

# Special object informing doc generation tooling which symbols
# to document & to associate with this module.
__docspec__ = [
    # Re-exported from trace.api
    init,
    publish,
    ref,
    require_current_call,
    get_current_call,
    finish,
    op,
    attributes,
    # Re-exported from flow module
    Object,
    Dataset,
    Model,
    Evaluation,
    Scorer,
]



import dataclasses
import typing
from typing import Optional, cast

import weave
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import Node, get, use
from weave.legacy.weave import context_state as _context
from weave.legacy.weave import panels
from weave.legacy.weave.ops_arrow import ArrowWeaveList, ArrowWeaveListType

_loading_builtins_token = _context.set_loading_built_ins()


@weave.type()
class RunSegment:
    run_name: str

    # this is an ArrowWeaveList containing a table with an arbitrary schema
    # except that one column should be called "step" and have cells of any
    # integer type
    metrics: typing.TypeVar("MetricRows")  # type: ignore

    prior_run_ref: typing.Optional[str] = dataclasses.field(
        default_factory=lambda: None
    )
    # index of the prior run's
    prior_run_branch_index: typing.Optional[int] = dataclasses.field(
        default_factory=lambda: None
    )

    @property
    def prior_run_branch_step(self) -> Optional[int]:
        if self.prior_run_branch_index is None:
            return None
        return self.metrics._index(self.prior_run_branch_index)["step"]

    def _experiment_body(self, limit: Optional[int] = None) -> ArrowWeaveList:

        if limit is None:
            limit = len(self.metrics)

        limited = self.metrics._limit(limit)._append_column(
            "run_name",
            [self.run_name] * limit,
        )

        if self.prior_run_ref is None:
            return limited

        # get the prior run
        prior_run: RunSegment = use(get(self.prior_run_ref))
        prior_experiment = prior_run._experiment_body(
            limit=self.prior_run_branch_index + 1
            if self.prior_run_branch_index is not None
            else 0
        )

        if len(prior_experiment) == 0:
            raise ValueError(
                f"Attempted to branch off of an empty run: run {prior_run} has no metrics."
            )

        return prior_experiment.concat(limited)

    @weave.op(render_info={"type": "function"}, hidden=True)
    def refine_experiment_type(self) -> types.Type:
        return ArrowWeaveListType(
            object_type=types.TypedDict(
                {**self.metrics.object_type.property_types, "run_name": types.String()}
            )
        )

    @weave.op(refine_output_type=refine_experiment_type)
    def experiment(self) -> typing.Any:
        return self._experiment_body()


@weave.op()
def run_segment_render(
    run_segment_node: Node[RunSegment],
) -> panels.Card:

    # All methods callable on X are callable on weave.Node[X], but
    # the types arent' setup properly, so cast to tell the type-checker
    # TODO: Fix!
    run_segment = cast(RunSegment, run_segment_node)

    return panels.Card(
        title=run_segment.run_name,
        subtitle="Weave Run Segment",
        content=[
            panels.CardTab(
                name="History",
                content=run_segment.metrics,
            ),
        ],
    )


_context.clear_loading_built_ins(_loading_builtins_token)



from typing import Any, Optional

from pydantic import (
    BaseModel,
    ConfigDict,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    model_validator,
)

from weave.trace.op import ObjectRef, Op
from weave.trace.vals import WeaveObject, pydantic_getattribute
from weave.trace.weave_client import get_ref


class Object(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None

    # Allow Op attributes
    model_config = ConfigDict(
        ignored_types=(Op,),
        arbitrary_types_allowed=True,
        protected_namespaces=(),
        extra="forbid",
    )

    __str__ = BaseModel.__repr__

    # This is a "wrap" validator meaning we can run our own logic before
    # and after the standard pydantic validation.
    @model_validator(mode="wrap")
    @classmethod
    def handle_relocatable_object(
        cls, v: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ) -> Any:
        if isinstance(v, ObjectRef):
            return v.get()
        if isinstance(v, WeaveObject):
            # This is a relocated object, so destructure it into a dictionary
            # so pydantic can validate it.
            keys = v._val.__dict__.keys()
            fields = {}
            for k in keys:
                if k.startswith("_"):
                    continue
                val = getattr(v, k)
                fields[k] = val

            # pydantic validation will construct a new pydantic object
            def is_ignored_type(v: type) -> bool:
                return isinstance(v, cls.model_config["ignored_types"])

            allowed_fields = {k: v for k, v in fields.items() if not is_ignored_type(v)}
            new_obj = handler(allowed_fields)
            for k, kv in fields.items():
                if is_ignored_type(kv):
                    new_obj.__dict__[k] = kv

            # transfer ref to new object
            # We can't attach a ref directly to pydantic objects yet.
            # TODO: fix this. I think dedupe may make it so the user data ends up
            #    working fine, but not setting a ref here will cause the client
            #    to do extra work.
            if isinstance(v, WeaveObject):
                ref = get_ref(v)
                new_obj.__dict__["ref"] = ref
            # return new_obj

            return new_obj
        return handler(v)


# Enable ref tracking for Weave.Object
# We could try to do this on BaseModel, but we haven't proven that's safe.
# So only Weave Objects will get ref tracking behavior for now.
Object.__getattribute__ = pydantic_getattribute  # type: ignore



import threading
from typing import TYPE_CHECKING, Optional

from weave.legacy.weave import context_state
from weave.trace.errors import WeaveInitError

if TYPE_CHECKING:
    from weave.trace.weave_client import WeaveClient

_global_weave_client: Optional["WeaveClient"] = None
lock = threading.Lock()


def set_weave_client_global(client: Optional["WeaveClient"]) -> None:
    global _global_weave_client

    # These outer guards are to avoid expensive lock acquisition
    if client is not None and _global_weave_client is None:
        with lock:
            if _global_weave_client is None:
                _global_weave_client = client

    elif client is None and _global_weave_client is not None:
        with lock:
            if _global_weave_client is not None:
                _global_weave_client = client


def set_weave_client_context(client: Optional["WeaveClient"]) -> None:
    context_state._graph_client.set(client)


def get_weave_client() -> Optional["WeaveClient"]:
    if (context_client := context_state._graph_client.get()) is not None:
        return context_client
    return _global_weave_client


def require_weave_client() -> "WeaveClient":
    if (client := get_weave_client()) is None:
        raise WeaveInitError("You must call `weave.init(<project_name>)` first")
    return client



"""Environment variables used by weave."""

# TODO: we should put all other env vars here to keep them organized.

import configparser
import enum
import json
import netrc
import os
import typing
from distutils.util import strtobool
from urllib.parse import urlparse
from weave.legacy.weave import errors
from weave.legacy.weave import util

if typing.TYPE_CHECKING:
    from weave.legacy.weave import logs

WANDB_ERROR_REPORTING = "WANDB_ERROR_REPORTING"
WEAVE_USAGE_ANALYTICS = "WEAVE_USAGE_ANALYTICS"
WEAVE_GQL_SCHEMA_PATH = "WEAVE_GQL_SCHEMA_PATH"


def _env_as_bool(var: str, default: typing.Optional[str] = None) -> bool:
    env = os.environ
    val = env.get(var, default)
    try:
        val = bool(strtobool(val))  # type: ignore
    except (AttributeError, ValueError):
        pass
    return val if isinstance(val, bool) else False


class Settings:
    """A minimal readonly implementation of wandb/old/settings.py for reading settings"""

    DEFAULT_SECTION = "default"
    DEFAULT_BASE_URL = "https://api.wandb.ai"

    def __init__(self) -> None:
        self._settings = configparser.ConfigParser()
        if not self._settings.has_section(self.DEFAULT_SECTION):
            self._settings.add_section(self.DEFAULT_SECTION)
        self._settings.read([Settings._global_path(), Settings._local_path()])

    @property
    def base_url(self) -> str:
        try:
            return self._settings.get(self.DEFAULT_SECTION, "base_url")
        except configparser.NoOptionError:
            return self.DEFAULT_BASE_URL

    @staticmethod
    def _local_path() -> str:
        return os.path.join(os.getcwd(), "wandb", "settings")

    @staticmethod
    def _global_path() -> str:
        default_config_dir = os.path.join(os.path.expanduser("~"), ".config", "wandb")
        config_dir = os.environ.get("WANDB_CONFIG_DIR", default_config_dir)
        return os.path.join(config_dir, "settings")


# There are currently two cache modes:
# - full: cache all cacheable intermediate results
# - minimal: cache only what we're sure we need to cache for performance
# WEAVE_NO_CACHE indicates that we're using the minimal cache mode. Otherwise
# we'll use full.
class CacheMode(enum.Enum):
    FULL = "full"
    MINIMAL = "minimal"


def cache_mode() -> CacheMode:
    if util.parse_boolean_env_var("WEAVE_NO_CACHE"):
        return CacheMode.MINIMAL
    env_mode = os.getenv("WEAVE_CACHE_MODE", CacheMode.MINIMAL.value)
    for mode in CacheMode:
        if mode.value == env_mode:
            return mode
    raise errors.WeaveConfigurationError(
        f"WEAVE_CACHE_MODE must be one of {list(CacheMode)}"
    )


# represents the number of days to re-use the weave cache before switching to a new directory
# a low number will cause frequent cache rotations and may lead to performance degradation due to
# cache misses. A high number will result in a large cache size which may cause infrastructure issues.
# if unset, the cache will write to a path that will not be cleaned up by the default cleanup task
def cache_duration_days() -> int:
    return int(os.getenv("WEAVE_CACHE_DURATION_DAYS", 0))


# represents the number of days after a cache has expired to leave it until deletion
def cache_deletion_buffer_days() -> int:
    return int(os.getenv("WEAVE_CACHE_DELETION_BUFFER_DAYS", 1))


def wandb_production() -> bool:
    return os.getenv("WEAVE_ENV") == "wandb_production"


def is_public() -> bool:
    return wandb_production()


def weave_log_format(default: "logs.LogFormat") -> "logs.LogFormat":
    from weave.legacy.weave.logs import LogFormat

    return LogFormat(os.getenv("WEAVE_LOG_FORMAT", default))


def weave_link_prefix() -> str:
    """When running in server we mount index under /weave"""
    if os.getenv("GORILLA_ONPREM") == "true":
        return "/weave"
    return ""


def weave_onprem() -> bool:
    return os.getenv("GORILLA_ONPREM") == "true"


def weave_backend_host() -> str:
    return os.getenv("WEAVE_BACKEND_HOST", "/__weave")


def trace_backend_base_url() -> str:
    return os.getenv("TRACE_BACKEND_BASE_URL", "")


def analytics_disabled() -> bool:
    if os.getenv("WEAVE_DISABLE_ANALYTICS") == "true":
        return True
    return False


def weave_server_url() -> str:
    base_url = wandb_frontend_base_url()
    default = "https://weave.wandb.ai"
    if base_url != "https://api.wandb.ai":
        default = base_url + "/weave"
    return os.getenv("WEAVE_SERVER_URL", default)


def weave_trace_server_url() -> str:
    base_url = wandb_frontend_base_url()
    default = "https://trace.wandb.ai"
    if base_url != "https://api.wandb.ai":
        default = base_url + "/traces"
    return os.getenv("WF_TRACE_SERVER_URL", default)


def wandb_base_url() -> str:
    settings = Settings()
    return os.environ.get("WANDB_BASE_URL", settings.base_url).rstrip("/")


def wandb_frontend_base_url() -> str:
    public_url = os.getenv("WANDB_PUBLIC_BASE_URL", "").rstrip("/")
    return public_url if public_url != "" else wandb_base_url()


# use filesystem.get_filesystem_dir() instead of directly accessing the env var
def weave_filesystem_dir() -> str:
    # WEAVE_LOCAL_ARTIFACT_DIR should be renamed to WEAVE_FILESYSTEM_DIR
    # TODO
    return os.environ.get("WEAVE_LOCAL_ARTIFACT_DIR") or os.path.join(
        "/tmp", "weave", "fs"
    )


def enable_touch_on_read() -> bool:
    return util.parse_boolean_env_var("WEAVE_ENABLE_TOUCH_ON_READ")


def memdump_sighandler_enabled() -> bool:
    return util.parse_boolean_env_var("WEAVE_ENABLE_MEMDUMP_SIGHANDLER")


def sigterm_sighandler_enabled() -> bool:
    return util.parse_boolean_env_var("WEAVE_ENABLE_SIGTERM_SIGHANDLER")


def weave_wandb_gql_headers() -> typing.Dict[str, str]:
    # expects a json string
    raw = os.environ.get("WEAVE_WANDB_GQL_HEADERS")
    if raw:
        try:
            return json.loads(raw)
        except json.JSONDecodeError:
            raise errors.WeaveConfigurationError(
                "WEAVE_WANDB_GQL_HEADERS should be a json string"
            )
    return {}


def weave_wandb_cookie() -> typing.Optional[str]:
    cookie = os.environ.get("WEAVE_WANDB_COOKIE")
    if cookie:
        if is_public():
            raise errors.WeaveConfigurationError(
                "WEAVE_WANDB_COOKIE should not be set in public mode."
            )
        for netrc_file in ("~/.netrc", "~/_netrc"):
            if os.path.exists(os.path.expanduser(netrc_file)):
                raise errors.WeaveConfigurationError(
                    f"Please delete {netrc_file} while using WEAVE_WANDB_COOKIE to avoid using your credentials"
                )
    return cookie


def stack_dump_sighandler_enabled() -> bool:
    return util.parse_boolean_env_var("WEAVE_ENABLE_STACK_DUMP_SIGHANDLER")


def _wandb_api_key_via_env() -> typing.Optional[str]:
    api_key = os.environ.get("WANDB_API_KEY")
    if api_key and is_public():
        raise errors.WeaveConfigurationError(
            "WANDB_API_KEY should not be set in public mode."
        )
    return api_key


def _wandb_api_key_via_netrc() -> typing.Optional[str]:
    for filepath in ("~/.netrc", "~/_netrc"):
        api_key = _wandb_api_key_via_netrc_file(filepath)
        if api_key:
            return api_key
    return None


def _wandb_api_key_via_netrc_file(filepath: str) -> typing.Optional[str]:
    netrc_path = os.path.expanduser(filepath)
    if not os.path.exists(netrc_path):
        return None
    nrc = netrc.netrc(netrc_path)
    res = nrc.authenticators(urlparse(wandb_base_url()).netloc)
    api_key = None
    if res:
        user, account, api_key = res
    if api_key and is_public():
        raise errors.WeaveConfigurationError(
            f"{filepath} should not be set in public mode."
        )
    return api_key


def weave_wandb_api_key() -> typing.Optional[str]:
    env_api_key = _wandb_api_key_via_env()
    netrc_api_key = _wandb_api_key_via_netrc()
    if env_api_key and netrc_api_key:
        if wandb_production():
            raise errors.WeaveConfigurationError(
                "WANDB_API_KEY should not be set in both ~/.netrc and the environment."
            )
        elif env_api_key != netrc_api_key:
            print(
                "There are different credentials in the netrc file and the environment. Using the environment value."
            )
    return env_api_key or netrc_api_key


def projection_timeout_sec() -> typing.Optional[typing.Union[int, float]]:
    return util.parse_number_env_var("WEAVE_PROJECTION_TIMEOUT_SEC")


def num_gql_timeout_retries() -> int:
    raw = util.parse_number_env_var("WEAVE_WANDB_GQL_NUM_TIMEOUT_RETRIES")
    if raw is None:
        return 0
    return int(raw)


def usage_analytics_enabled() -> bool:
    return _env_as_bool(WANDB_ERROR_REPORTING, default="True") and _env_as_bool(
        WEAVE_USAGE_ANALYTICS, default="True"
    )


def gql_schema_path() -> typing.Optional[str]:
    return os.environ.get(WEAVE_GQL_SCHEMA_PATH) or None


def dd_env() -> str:
    return os.getenv("DD_ENV", "")


def env_is_ci() -> bool:
    return util.parse_boolean_env_var("WEAVE_CI")


def disable_weave_pii() -> bool:
    return os.getenv("DISABLE_WEAVE_PII", "false").strip().lower() == "true"



# TODO: move all panelbank stuff here

from weave.legacy.weave.panels import panel_group


def default_panel_bank_flow_section_config():
    return panel_group.PanelBankSectionConfig(
        id="grid0",
        name="Section 0",
        panels=[],
        isOpen=True,
        flowConfig=panel_group.PanelBankFlowSectionConfig(
            snapToColumns=True,
            columnsPerPage=1,
            rowsPerPage=1,
            gutterWidth=0,
            boxWidth=64,
            boxHeight=64,
        ),
        type="grid",
        sorted=0,
    )


def flow_layout(rows: int, cols: int):
    layout = default_panel_bank_flow_section_config()
    layout["flowConfig"]["columnsPerPage"] = cols
    layout["flowConfig"]["rowsPerPage"] = rows
    return layout



import datetime
import uuid

from weave.trace import weave_client
from weave.trace_server import trace_server_interface as tsi


def test_call_update_out_of_order(client: weave_client.WeaveClient):
    # Here, we are going to do an out of order sequence:
    # 1. Name a call
    # 2. Finish the call
    # 3. Start the call
    # 4. Delete the call
    #
    # Only step 3 should actually have the result in the query.

    call_id = str(uuid.uuid4())
    project_id = client._project_id()

    def get_calls():
        res = client.server.calls_query(
            tsi.CallsQueryReq(
                project_id=project_id,
            )
        )
        return res.calls

    client.server.call_update(
        tsi.CallUpdateReq(
            project_id=project_id,
            call_id=call_id,
            display_name="test_display_name",
        )
    )

    assert len(get_calls()) == 0

    client.server.call_end(
        tsi.CallEndReq(
            end=tsi.EndedCallSchemaForInsert(
                project_id=project_id,
                id=call_id,
                ended_at=datetime.datetime.now(),
                exception=None,
                output=None,
                summary={},
            )
        )
    )

    assert len(get_calls()) == 0

    client.server.call_start(
        tsi.CallStartReq(
            start=tsi.StartedCallSchemaForInsert(
                project_id=project_id,
                id=call_id,
                trace_id=call_id,
                started_at=datetime.datetime.now(),
                op_name="test_op_name",
                attributes={},
                inputs={},
            )
        )
    )

    assert len(get_calls()) == 1

    client.server.calls_delete(
        tsi.CallsDeleteReq(
            project_id=project_id,
            call_ids=[call_id],
        )
    )

    assert len(get_calls()) == 0



import typing

import weave
from weave.legacy.weave import registry_mem
from weave.legacy.weave.ops_domain import run_ops, wb_domain_types

# We can't chain ops called .name() because of a weird bug :( [its a field on VarNode].
# So we have to get the ops here and call them directly for now.
project_name_op = registry_mem.memory_registry.get_op("project-name")
entity_name_op = registry_mem.memory_registry.get_op("entity-name")
user_name_op = registry_mem.memory_registry.get_op("user-name")


@weave.type()
class FakeWandbModel:
    name: str


@weave.op(
    render_info={"type": "function"},
)
def org_model(entity_name: str, model_name: str) -> FakeWandbModel:
    return FakeWandbModel(model_name)


GHOSTWRITE_MD = weave.legacy.weave.ops.Markdown(
    """
# [ghostwrite.ai](https://ghostwrite.ai).

Ghostwrite is a tool for writing. Our mission is to increase human creativity.
"""
)


@weave.op()
def fakewandbmodel_render(
    model_node: weave.Node[FakeWandbModel],
) -> weave.legacy.weave.panels.Card:
    model = typing.cast(FakeWandbModel, model_node)
    return weave.legacy.weave.panels.Card(
        title=model.name,
        subtitle="",
        content=[
            weave.legacy.weave.panels.CardTab(
                name="Description",
                content=weave.legacy.weave.panels.PanelMarkdown(GHOSTWRITE_MD),  # type: ignore
            ),
            weave.legacy.weave.panels.CardTab(
                name="Predictions",
                content=weave.legacy.weave.ops.project("shawn", "ghostwrite-test1")
                .runs()
                .summary()["predictions"]
                .table()
                .rows()
                .concat()
                .to_py(),
            ),
        ],
    )


@weave.op()
def entity_render(
    entity_node: weave.Node[wb_domain_types.Entity],
) -> weave.legacy.weave.panels.Card:
    entity = typing.cast(wb_domain_types.Entity, entity_node)
    return weave.legacy.weave.panels.Card(
        title=entity_name_op(entity),
        subtitle="",
        content=[
            weave.legacy.weave.panels.CardTab(
                name="Projects",
                content=weave.legacy.weave.panels.Table(
                    entity.projects(),  # type: ignore
                    columns=[
                        lambda project: weave.legacy.weave.panels.WeaveLink(
                            project_name_op(project),
                            vars={
                                "entity_name": entity_name_op(project.entity()),
                                "project_name": project_name_op(project),
                            },
                            to=lambda input, vars: weave.legacy.weave.ops.project(
                                vars["entity_name"], vars["project_name"]
                            ),
                        ),
                    ],
                ),
            ),
            weave.legacy.weave.panels.CardTab(
                name="Registered Models",
                content=weave.legacy.weave.panels.Table(
                    weave.save(
                        ["ghostwrite", "credit card predictor", "stopsigns3"],
                        name="model_list",
                    ),
                    columns=[
                        lambda model_name: weave.legacy.weave.panels.WeaveLink(
                            model_name,
                            vars={
                                "entity_name": entity_name_op(entity),
                            },
                            to=lambda input, vars: org_model(
                                vars["entity_name"], input
                            ),
                        ),
                    ],
                ),
            ),
        ],
    )


@weave.type()
class ProjectRunsTable(weave.Panel):
    id = "project-runs-table"
    input_node: weave.Node[list[wb_domain_types.Run]]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.Table:
        return weave.legacy.weave.panels.Table(
            self.input_node,
            columns=[
                lambda run: weave.legacy.weave.panels.WeaveLink(
                    run.id(),
                    vars={
                        "entity_name": entity_name_op(run.project().entity()),
                        "project_name": project_name_op(run.project()),
                        "run_id": run.id(),
                    },
                    to=lambda input, vars: weave.legacy.weave.ops.project(
                        vars["entity_name"], vars["project_name"]
                    ).run(vars["run_id"]),
                ),
                lambda run: run_ops.run_name(run),
                lambda run: run.createdAt(),
            ],
        )


@weave.type()
class ProjectArtifactsTable(weave.Panel):
    id = "project-artifacts-table"
    input_node: weave.Node[list[wb_domain_types.ArtifactCollection]]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.Table:
        # This breaks if there is a variable in the node
        # types_names = weave.use(artifacts._get_op("type")().name().unique())
        # return weave.legacy.weave.panels.Card(
        #     title="Artifacts",
        #     subtitle="",
        #     content=[
        #         weave.legacy.weave.panels.CardTab(name=type_name, content=[type_name])
        #         for type_name in types_names
        #     ],
        # )
        return weave.legacy.weave.panels.Table(
            self.input_node,
            columns=[
                lambda artifact: weave.legacy.weave.panels.WeaveLink(
                    artifact._get_op("name")(),
                    vars={
                        "entity_name": entity_name_op(artifact.project().entity()),
                        "project_name": project_name_op(artifact.project()),
                        "artifact_name": artifact._get_op("name")(),
                    },
                    to=lambda input, vars: weave.legacy.weave.ops.project(
                        vars["entity_name"], vars["project_name"]
                    ).artifact(vars["artifact_name"]),
                ),
                lambda artifact: artifact.createdAt(),
            ],
        )



import datetime
import inspect
import typing
from typing import Optional

from fastapi import Depends, FastAPI, Header, HTTPException
from fastapi.security import HTTPBasic, HTTPBasicCredentials

try:
    from typing import Annotated
# Support python 3.8
except ImportError:
    from typing_extensions import Annotated  # type: ignore

from weave.legacy.weave import cache, op_args, pyfunc_type_util, weave_pydantic
from weave.legacy.weave.wandb_api import WandbApiAsync
from weave.trace import errors
from weave.trace.op import Op, is_op
from weave.trace.refs import ObjectRef

key_cache: cache.LruTimeWindowCache[str, typing.Optional[bool]] = (
    cache.LruTimeWindowCache(datetime.timedelta(minutes=5))
)

api: Optional[WandbApiAsync] = None


def wandb_auth(
    entity: str,
) -> typing.Callable[
    [typing.Optional[str]], typing.Coroutine[typing.Any, typing.Any, bool]
]:
    async def auth_inner(key: Annotated[Optional[str], Depends(api_key)]) -> bool:
        global api
        if api is None:
            api = WandbApiAsync()
        if key is None:
            raise HTTPException(status_code=401, detail="Missing API Key")
        if len(key.split("-")[-1]) != 40:
            raise HTTPException(status_code=401, detail="Invalid API Key")

        authed = key_cache.get(key)
        if isinstance(authed, bool):
            return authed
        authed = await api.can_access_entity(entity, api_key=key)
        if not authed:
            raise HTTPException(status_code=403, detail="Permission Denied")
        key_cache.set(key, authed)
        return authed

    return auth_inner


def api_key(
    credentials: Annotated[
        Optional[HTTPBasicCredentials],
        Depends(
            HTTPBasic(
                auto_error=False,
                description="Set your username to api and password to a W&B API Key",
            )
        ),
    ],
    x_wandb_api_key: Annotated[
        Optional[str], Header(description="Optional W&B API Key")
    ] = None,
) -> Optional[str]:
    if x_wandb_api_key:
        return x_wandb_api_key
    elif credentials and credentials.password:
        return credentials.password
    else:
        return None


def object_method_app(
    obj_ref: ObjectRef,
    method_name: typing.Optional[str] = None,
    auth_entity: typing.Optional[str] = None,
) -> FastAPI:
    obj = obj_ref.get()

    attrs: dict[str, Op] = {attr: getattr(obj, attr) for attr in dir(obj)}
    op_attrs = {k: v for k, v in attrs.items() if is_op(v)}

    if not op_attrs:
        raise ValueError("No ops found on object")

    if method_name is None:
        if len(op_attrs) > 1:
            raise ValueError(
                "Multiple ops found on object (%s), must specify method_name argument"
                % ", ".join(op_attrs)
            )
        method_name = next(iter(op_attrs))

    if (method := getattr(obj, method_name, None)) is None:
        raise ValueError(f"Method {method_name} not found")

    if not is_op(unbound_method := method.__func__):
        raise ValueError(f"Expected an op, got {unbound_method}")

    try:
        args = pyfunc_type_util.determine_input_type(unbound_method)
    except errors.WeaveDefinitionError as e:
        raise ValueError(
            f"Type for model's method '{method_name}' could not be determined. Did you annotate it with Python types? {e}"
        )
    if not isinstance(args, op_args.OpNamedArgs):
        raise ValueError("predict op must have named args")

    arg_types = args.weave_type().property_types
    del arg_types["self"]

    Item = weave_pydantic.weave_type_to_pydantic(arg_types, name="Item")

    dependencies = []
    if auth_entity:
        dependencies.append(Depends(wandb_auth(auth_entity)))

    app = FastAPI(dependencies=dependencies)

    @app.post(f"/{method_name}", summary=method_name)
    async def method_route(item: Item) -> dict:  # type: ignore
        if inspect.iscoroutinefunction(method):
            result = await method(**item.dict())  # type: ignore
        else:
            result = method(**item.dict())  # type: ignore
        return {"result": result}

    return app



import os
from typing import Any

import pytest

import weave
from weave.trace_server import trace_server_interface as tsi


def _get_call_output(call: tsi.CallSchema) -> Any:
    """This is a hack and should not be needed. We should be able to auto-resolve this for the user.

    Keeping this here for now, but it should be removed in the future once we have a better solution.
    """
    call_output = call.output
    if isinstance(call_output, str) and call_output.startswith("weave://"):
        return weave.ref(call_output).get()
    return call_output


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_mistral_quickstart(client: weave.trace.weave_client.WeaveClient) -> None:
    # This is taken directly from https://docs.mistral.ai/getting-started/quickstart/
    from mistralai import Mistral  # type: ignore

    api_key = os.environ.get("MISTRAL_API_KEY", "DUMMY_API_KEY")
    model = "mistral-large-latest"

    mistral_client = Mistral(api_key=api_key)

    chat_response = mistral_client.chat.complete(
        model=model,
        messages=[{"role": "user", "content": "What is the best French cheese?"}],
    )

    all_content = chat_response.choices[0].message.content
    exp = """Choosing the "best" French cheese can be subjective, as it largely depends on personal taste. France is renowned for its wide variety of cheeses, with over 400 different types. Here are a few highly regarded French cheeses across various categories:

1. **Soft Cheeses**:
   - **Brie de Meaux**: Known for its creamy texture and rich, buttery flavor.
   - **Camembert de Normandie**: Soft, creamy, and has a distinctive earthy aroma.

2. **Semi-Soft Cheeses**:
   - **Morbier**: Recognizable by the layer of ash in the middle, it has a fruity and slightly nutty taste.
   - **Reblochon**: A savory cheese from the Alps, often used in tartiflette, a traditional dish.

3. **Hard Cheeses**:
   - **Comté**: A nutty and slightly sweet cheese, similar to Gruyère.
   - **Beaufort**: Known for its firm texture and complex, nutty flavor.

4. **Blue Cheeses**:
   - **Roquefort**: A tangy and pungent blue cheese made from sheep's milk.
   - **Bleu d'Auvergne**: A creamy and strong-flavored blue cheese.

5. **Goat Cheeses**:
   - **Chèvre**: Available in many forms, from fresh and creamy to aged and crumbly.
   - **Crottin de Chavignol**: A small, round goat cheese with a nutty flavor.

Each of these cheeses has its unique characteristics, so the "best" one depends on your preferences. It's always fun to try several and decide which you like the most!"""
    assert all_content == exp
    res = client.server.calls_query(tsi.CallsQueryReq(project_id=client._project_id()))
    assert len(res.calls) == 1
    call = res.calls[0]
    assert call.exception is None and call.ended_at is not None
    output = _get_call_output(call)
    print(f"{output['choices'][0]=}")
    assert output["choices"][0]["message"]["content"] == exp
    assert output["choices"][0]["finish_reason"] == "stop"
    assert output["id"] == chat_response.id
    assert output["model"] == chat_response.model
    assert output["object"] == chat_response.object
    assert output["created"] == chat_response.created
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output["model"]]
    assert model_usage["requests"] == 1
    assert (
        output["usage"]["completion_tokens"] == model_usage["completion_tokens"] == 406
    )
    assert output["usage"]["prompt_tokens"] == model_usage["prompt_tokens"] == 10
    assert output["usage"]["total_tokens"] == model_usage["total_tokens"] == 416


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_mistral_quickstart_async(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from mistralai import Mistral  # type: ignore

    # This is taken directly from https://docs.mistral.ai/getting-started/quickstart/
    api_key = os.environ.get("MISTRAL_API_KEY", "DUMMY_API_KEY")
    model = "mistral-large-latest"

    mistral_client = Mistral(api_key=api_key)

    chat_response = await mistral_client.chat.complete_async(
        model=model,
        messages=[{"role": "user", "content": "What is the best French cheese?"}],
    )

    all_content = chat_response.choices[0].message.content
    exp = """Choosing the "best" French cheese is quite subjective and depends on personal taste, as France has a wide variety of excellent cheeses. Here are a few renowned ones that are often considered among the best:

1. **Roquefort**: This is a sheep milk blue cheese from the south of France. It's known for its tangy and salty flavor, and distinctive veins of blue mold.

2. **Camembert de Normandie**: A soft, creamy, surface-ripened cow's milk cheese. It has a rich, buttery flavor and is often enjoyed when fully ripe and gooey.

3. **Brie de Meaux**: Another soft cow's milk cheese, similar to Camembert but slightly milder in flavor. It's known for its edible white rind and creamy interior.

4. **Comté**: A French cheese made from unpasteurized cow's milk in the Franche-Comté region. It has a complex, nutty flavor and a firm, springy texture.

5. **Reblochon**: A soft, washed-rind and smear-ripened cheese made in the Alpine region of Savoy from raw cow's milk. It has a nutty taste and a strong aroma.

6. **Époisses de Bourgogne**: A pungent, washed-rind cheese made from cow's milk. It has a strong, somewhat salty flavor and a sticky, orange exterior.

Each of these cheeses offers a unique taste and texture, so the "best" one is a matter of personal preference."""

    assert all_content == exp
    res = client.server.calls_query(tsi.CallsQueryReq(project_id=client._project_id()))
    assert len(res.calls) == 1
    call = res.calls[0]
    assert call.exception is None and call.ended_at is not None
    output = _get_call_output(call)
    assert output["choices"][0]["message"]["content"] == exp
    assert output["choices"][0]["finish_reason"] == "stop"
    assert output["id"] == chat_response.id
    assert output["model"] == chat_response.model
    assert output["object"] == chat_response.object
    assert output["created"] == chat_response.created
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output["model"]]
    assert model_usage["requests"] == 1
    assert (
        output["usage"]["completion_tokens"] == model_usage["completion_tokens"] == 363
    )
    assert output["usage"]["prompt_tokens"] == model_usage["prompt_tokens"] == 10
    assert output["usage"]["total_tokens"] == model_usage["total_tokens"] == 373


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_mistral_quickstart_with_stream(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from mistralai import Mistral  # type: ignore

    # This is taken directly from https://docs.mistral.ai/getting-started/quickstart/
    api_key = os.environ.get("MISTRAL_API_KEY", "DUMMY_API_KEY")
    model = "mistral-large-latest"

    mistral_client = Mistral(api_key=api_key)

    chat_response = mistral_client.chat.stream(
        model=model,
        messages=[{"role": "user", "content": "What is the best French cheese?"}],
    )

    all_content = ""
    for chunk in chat_response:
        all_content += chunk.data.choices[0].delta.content

    exp = """Choosing the "best" French cheese can be subjective as it depends on personal taste, but France is renowned for its wide variety of high-quality cheeses. Here are a few that are often highly regarded:

1. **Roquefort**: A sheep milk blue cheese from the south of France, known for its tangy, salty flavor and distinctive veins of blue mold.

2. **Brie de Meaux**: A soft, creamy cow's milk cheese from the Île-de-France region. It has a rich, buttery flavor and is often considered one of the finest cheeses in the world.

3. **Camembert de Normandie**: A soft, creamy cow's milk cheese from Normandy, similar to Brie but with a slightly stronger flavor.

4. **Comté**: A hard, unpasteurized cow's milk cheese from the Jura Massif region. It has a complex, nutty flavor that varies depending on its age.

5. **Époisses de Bourgogne**: A pungent, soft cow's milk cheese from Burgundy, often served with a spoon due to its runny texture.

6. **Morbier**: A semi-soft cow's milk cheese with a distinctive layer of ash in the middle. It has a rich, creamy flavor with a slight bitterness.

Each of these cheeses offers a unique taste and texture, so the "best" one depends on your personal preference. It's always fun to try a variety to see which you like best!"""
    assert all_content == exp
    res = client.server.calls_query(tsi.CallsQueryReq(project_id=client._project_id()))
    assert len(res.calls) == 1
    call = res.calls[0]
    assert call.exception is None and call.ended_at is not None
    output = _get_call_output(call)
    print(f"{output=}")
    assert output["choices"][0]["message"]["content"] == exp
    assert output["choices"][0]["finish_reason"] == "stop"
    assert output["id"] == chunk.data.id
    assert output["model"] == chunk.data.model
    assert output["object"] == chunk.data.object
    assert output["created"] == chunk.data.created
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output["model"]]
    assert model_usage["requests"] == 1
    assert (
        output["usage"]["completion_tokens"] == model_usage["completion_tokens"] == 350
    )
    assert output["usage"]["prompt_tokens"] == model_usage["prompt_tokens"] == 10
    assert output["usage"]["total_tokens"] == model_usage["total_tokens"] == 360


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_mistral_quickstart_with_stream_async(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from mistralai import Mistral  # type: ignore

    # This is taken directly from https://docs.mistral.ai/getting-started/quickstart/
    api_key = os.environ.get("MISTRAL_API_KEY", "DUMMY_API_KEY")
    model = "mistral-large-latest"

    mistral_client = Mistral(api_key=api_key)

    chat_response = await mistral_client.chat.stream_async(
        model=model,
        messages=[{"role": "user", "content": "What is the best French cheese?"}],
    )

    all_content = ""
    async for chunk in chat_response:
        all_content += chunk.data.choices[0].delta.content

    exp = """Choosing the "best" French cheese can be subjective as it largely depends on personal taste. France is renowned for its wide variety of cheeses, with over 400 different types. Here are a few highly regarded French cheeses across various categories:

1. **Soft Cheeses**:
   - **Brie de Meaux**: Known for its rich, creamy texture and earthy, mushroom-like flavor.
   - **Camembert de Normandie**: Soft, creamy, and has a strong, distinctive smell.

2. **Semi-Soft Cheeses**:
   - **Morbier**: Recognizable by its layer of ash in the middle, it has a fruity and slightly nutty taste.
   - **Reblochon**: A savory cheese from the Alps with a nutty aftertaste and a soft, washed rind.

3. **Hard Cheeses**:
   - **Comté**: A fruity and nutty cheese made from unpasteurized cow's milk, often aged for many months.
   - **Beaufort**: Similar to Gruyère, it has a firm texture and a sweet, nutty flavor.

4. **Blue Cheeses**:
   - **Roquefort**: A sheep milk cheese with distinctive veins of blue mold, offering a tangy and salty taste.
   - **Bleu d'Auvergne**: A cow's milk blue cheese with a strong, pungent aroma and a creamy texture.

5. **Goat Cheeses**:
   - **Chèvre**: Available in many forms, from fresh and creamy to aged and crumbly, often with a tangy, earthy flavor.
   - **Crottin de Chavignol**: A small, round goat cheese with a nutty flavor and a texture that varies with age.

Each of these cheeses has its unique characteristics, so the "best" one depends on your preferences. It's always fun to try several types to discover your favorite!"""
    assert all_content == exp
    res = client.server.calls_query(tsi.CallsQueryReq(project_id=client._project_id()))
    assert len(res.calls) == 1
    call = res.calls[0]
    assert call.exception is None and call.ended_at is not None
    output = _get_call_output(call)
    assert output["choices"][0]["message"]["content"] == exp
    assert output["choices"][0]["finish_reason"] == "stop"
    assert output["id"] == chunk.data.id
    assert output["model"] == chunk.data.model
    assert output["object"] == chunk.data.object
    assert output["created"] == chunk.data.created
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output["model"]]
    assert model_usage["requests"] == 1
    assert (
        output["usage"]["completion_tokens"] == model_usage["completion_tokens"] == 459
    )
    assert output["usage"]["prompt_tokens"] == model_usage["prompt_tokens"] == 10
    assert output["usage"]["total_tokens"] == model_usage["total_tokens"] == 469



import typing

import pytest

from weave.legacy.weave import api as weave
from weave.legacy.weave import context_state, graph, storage, uris, weave_internal
from weave.legacy.weave import context_state as _context_state
from weave.legacy.weave import weave_types as types

from . import test_helpers

_loading_builtins_token = _context_state.set_loading_built_ins()


@weave.op(
    name="test_op-op_simple",
    input_type={"a": types.Int(), "b": types.Int()},
    output_type=types.String(),
)
def op_simple(a, b):
    return str(a) + str(b)


def test_op_simple():
    x = op_simple(3, 4)
    assert str(x) == "op_simple(3, 4)"
    test_helpers.assert_nodes_equal(
        x,
        graph.OutputNode(
            types.String(),
            test_helpers.RegexMatcher(".*test_op-op_simple.*"),
            {
                "a": weave_internal.make_const_node(types.Int(), 3),
                "b": weave_internal.make_const_node(types.Int(), 4),
            },
        ),
    )
    assert weave.use(x) == "34"


@weave.op()
def op_inferredtype(a: int, b: int) -> str:
    return str(a) + str(b)


_context_state.clear_loading_built_ins(_loading_builtins_token)


def test_op_inferred_type():
    assert op_inferredtype.input_type.arg_types == {
        "a": types.Int(),
        "b": types.Int(),
    }
    assert op_inferredtype.concrete_output_type == types.String()


def test_op_incompatible_return_type():
    with pytest.raises(weave.errors.WeaveDefinitionError):
        _t = _context_state.set_loading_built_ins()

        try:

            @weave.op(output_type=types.Int())
            def op_invalid_returntype(a: int, b: int) -> str:
                return str(a) + str(b)

        finally:
            _context_state.clear_loading_built_ins(_t)


def test_op_incompatible_input_arg_type():
    with pytest.raises(weave.errors.WeaveDefinitionError):
        _t = _context_state.set_loading_built_ins()

        try:

            @weave.op(
                input_type={"a": types.String(), "b": types.Int()},
                output_type=types.Int(),
            )
            def op_invalid_input_arg_type(a: int, b: int) -> str:
                return str(a) + str(b)

        finally:
            _context_state.clear_loading_built_ins(_t)


def test_op_too_few_input_arg_type():
    with pytest.raises(weave.errors.WeaveDefinitionError):
        _t = _context_state.set_loading_built_ins()

        try:

            @weave.op(input_type={"a": types.Int()}, output_type=types.Int())
            def op_invalid_too_few(a: int, b: int) -> str:
                return str(a) + str(b)

        finally:
            _context_state.clear_loading_built_ins(_t)


def test_op_too_many_input_arg_type():
    with pytest.raises(weave.errors.WeaveDefinitionError):
        _t = _context_state.set_loading_built_ins()

        try:

            @weave.op(
                input_type={"a": types.Int(), "b": types.Int(), "c": types.Int()},
            )
            def op_invalid_too_many(a: int, b: int) -> str:
                return str(a) + str(b)

        finally:
            _context_state.clear_loading_built_ins(_t)


def test_op_callable_output_type_and_return_type_declared():
    with pytest.raises(weave.errors.WeaveDefinitionError):
        _t = _context_state.set_loading_built_ins()
        try:

            @weave.op(
                input_type={"a": types.Int()}, output_type=lambda a: types.String()
            )
            def op_callable_output_type_and_return_type_declared(a: int) -> str:
                return str(a)

        finally:
            _context_state.clear_loading_built_ins(_t)


def test_op_no_arg_type():
    with pytest.raises(weave.errors.WeaveDefinitionError):
        with context_state.loading_builtins(True):

            @weave.op()
            def op_callable_output_type_and_return_type_declared(a: int):
                return str(a)


class SomeUnknownObj:
    pass


@pytest.mark.skip("We allow Unknown types now")
def test_op_unknown_arg_type():
    with pytest.raises(weave.errors.WeaveDefinitionError):

        @weave.op()
        def op_callable_output_type_and_return_type_declared(a: SomeUnknownObj):
            return str(a)


def test_op_no_return_type():
    with pytest.raises(weave.errors.WeaveDefinitionError):
        with context_state.loading_builtins(True):

            @weave.op(input_type={"a": types.Int()})
            def op_callable_output_type_and_return_type_declared(a: int):
                return str(a)


def test_op_inferred_list_return():
    _t = _context_state.set_loading_built_ins()
    try:

        @weave.op()
        def op_under_test(a: int) -> list[int]:
            return [a, 2 * a, 3 * a]

    finally:
        _context_state.clear_loading_built_ins(_t)

    assert op_under_test.concrete_output_type == types.List(types.Int())


def test_op_inferred_typeddict_return():
    _t = _context_state.set_loading_built_ins()
    try:

        @weave.op()
        def op_test_op_inferred_typeddict_return_op(
            a: int,
        ) -> typing.TypedDict("OpReturn", {"x": int, "y": str}):
            return {"a": 1, "y": "x"}

    finally:
        _context_state.clear_loading_built_ins(_t)

    assert (
        op_test_op_inferred_typeddict_return_op.concrete_output_type
        == types.TypedDict({"x": types.Int(), "y": types.String()})
    )


def test_op_inferred_list_typeddict_return():
    _t = _context_state.set_loading_built_ins()
    try:

        @weave.op()
        def op_inferred_list_typeddict_return(
            a: int,
        ) -> list[typing.TypedDict("OpReturn", {"x": int, "y": str})]:
            return [{"a": 1, "y": "x"}]

    finally:
        _context_state.clear_loading_built_ins(_t)

    assert op_inferred_list_typeddict_return.concrete_output_type == types.List(
        types.TypedDict({"x": types.Int(), "y": types.String()})
    )


def test_op_inferred_dict_return() -> None:
    _t = _context_state.set_loading_built_ins()
    try:

        @weave.op()
        def op_inferred_dict_return(a: int) -> dict[str, list[int]]:
            return {"a": [5]}

    finally:
        _context_state.clear_loading_built_ins(_t)

    assert op_inferred_dict_return.concrete_output_type == types.Dict(
        types.String(), types.List(types.Int())
    )


def test_op_method_inferred_self():
    class SomeWeaveType(types.Type):
        name = "SomeWeaveType"

    _t = _context_state.set_loading_built_ins()
    try:

        @weave.weave_class(weave_type=SomeWeaveType)
        class SomeWeaveObj:
            @weave.op()
            def my_op(self, a: int) -> str:  # type: ignore[empty-body]
                pass

    finally:
        _context_state.clear_loading_built_ins(_t)

    assert SomeWeaveObj.my_op.input_type.arg_types == {
        "self": SomeWeaveType(),
        "a": types.Int(),
    }
    assert SomeWeaveObj.my_op.concrete_output_type == types.String()


def test_op_internal_tracing_enabled(client):
    # This test verifies the behavior of `_tracing_enabled` which
    # is not a user-facing API and is used internally to toggle
    # tracing on and off.
    @weave.op
    def my_op():
        return "hello"

    my_op()  # <-- this call will be traced

    assert len(list(my_op.calls())) == 1

    my_op._tracing_enabled = False

    my_op()  # <-- this call will not be traced

    assert len(list(my_op.calls())) == 1

    my_op._tracing_enabled = True

    my_op()  # <-- this call will be traced

    assert len(list(my_op.calls())) == 2



import gql

from weave.legacy.weave import wandb_api

ARTIFACT_LINEAGE_QUERY = gql.gql(
    """
    query ArtifactLineage(
        $entityName: String!,
        $projectName: String!,
        $artifactName: String!,
    ) {
        project(name: $projectName, entityName: $entityName) {
            artifact(name: $artifactName) {
                artifactLineageDag(limit: 10000, filterGeneratedArtifacts: true) {
                    artifacts {
                        artifactNodeID
                        entityName
                        projectName
                        artifactTypeName
                        artifactSequenceName
                        artifactCommitHash
                        versionIndex
                    }
                    runs {
                        runNodeID
                        entityName
                        projectName
                        runName
                        displayName
                        jobType
                    }
                    edges {
                        edgeID
                        artifactNodeID
                        runNodeID
                        direction
                    }
                    hitLimit
                }
            }
        }
    }
    """
)


def get_run_checkpoint_chain(entity: str, project: str, leaf_checkpoint_name: str):
    api = wandb_api.WandbApi()

    query_res = api.query(
        ARTIFACT_LINEAGE_QUERY,
        entityName=entity,
        projectName=project,
        artifactName=leaf_checkpoint_name,
    )

    dag = query_res["project"]["artifact"]["artifactLineageDag"]

    artifacts_dict = {a["artifactNodeID"]: a for a in dag["artifacts"]}
    runs_dict = {r["runNodeID"]: r for r in dag["runs"]}

    target_artifact_id = [
        a["artifactNodeID"]
        for a in dag["artifacts"]
        if f"{a['artifactSequenceName']}:v{a['versionIndex']}" == leaf_checkpoint_name
    ][0]

    # Create an empty sequence list.
    sequence = []

    # Traverse the DAG from the target artifact.
    current_artifact_id = target_artifact_id

    while True:
        # Find edges where the artifact is the target and the direction is 'AwayFromArtifact'.
        next_edges = [
            e
            for e in dag["edges"]
            if e["artifactNodeID"] == current_artifact_id
            and e["direction"] == "TowardArtifact"
        ]

        # If there are no such edges, we've reached the beginning of the lineage.
        if not next_edges:
            break

        # Get the corresponding run for the current artifact.
        current_run_id = next_edges[0]["runNodeID"]
        sequence.append(
            (runs_dict[current_run_id], artifacts_dict[current_artifact_id])
        )

        # Find the next artifact(s) where the run is the source and the direction is 'AwayFromArtifact'.
        next_edges = [
            e
            for e in dag["edges"]
            if e["runNodeID"] == current_run_id and e["direction"] == "AwayFromArtifact"
        ]

        # Find the checkpoint artifact among the next artifacts.
        checkpoint_artifacts = [
            e
            for e in next_edges
            if artifacts_dict[e["artifactNodeID"]]["artifactTypeName"] == "checkpoint"
        ]

        # If there's no checkpoint artifact, we've reached the beginning of the lineage.
        if not checkpoint_artifacts:
            break

        # Set the next artifact to the checkpoint artifact (since there should be only one checkpoint artifact, get the first).
        current_artifact_id = checkpoint_artifacts[0]["artifactNodeID"]

    return sequence



from weave.legacy.weave import api as weave
from weave.legacy.weave import storage
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.decorator_op import op


def test_function_op_name():
    @op()
    def test_decorators_function_op(a: int, b: int) -> int:
        return a + b

    assert test_decorators_function_op.name == "op-test_decorators_function_op"


def test_method_op_name():
    class MyObjType(types.Type):
        name = "test_decorators_my_obj"

    @weave.weave_class(weave_type=MyObjType)
    class MyObj:
        @op()
        def my_op(self: int, b: int) -> int:  # type: ignore
            return self + b

    assert MyObj.my_op.name == "test_decorators_my_obj-my_op"


@weave.type()
class Point:
    x: float
    y: float


def test_attr_access():
    p = weave.save(Point(1, 2), "my-point")
    assert weave.use(p.x) == 1
    assert weave.use(p.y) == 2


@weave.type()
class _TestDecoratorLineSegment:
    start: Point
    end: Point

    def mid(self):
        return Point(
            self.start.x + (self.end.x - self.start.x) / 2,
            self.start.y + (self.end.y - self.start.y) / 2,
        )


def test_weave_obj():
    box = Point(1, 2)
    box_type = weave.type_of(box)
    assert box_type == Point.WeaveType()
    ref = storage.save(box)
    box2 = ref.get()
    assert box == box2


def test_nested_weave_obj():
    line = _TestDecoratorLineSegment(Point(1, 2), Point(3, 4))
    line_type = weave.type_of(line)
    assert line_type == _TestDecoratorLineSegment.WeaveType()
    ref = storage.save(line)
    line2 = ref.get()
    assert line == line2


def make_lines() -> list[_TestDecoratorLineSegment]:
    lines = []
    for i in range(2):
        line = _TestDecoratorLineSegment(
            Point(float(i), float(i + 2)),
            Point(float(i * 2), float(i * 2 + 1)),
        )
        lines.append(line)
    return lines


def test_list_nested_weave_obj():
    lines = make_lines()
    lines_type = weave.type_of(lines)
    assert lines_type == types.List(_TestDecoratorLineSegment.WeaveType())
    ref = storage.save(lines)
    lines2 = ref.get()
    assert lines2 == lines



# Disable warning spew.
import warnings

warnings.filterwarnings("ignore", message=".*The 'nopython' keyword.*")

import pickle
import random
import typing

import matplotlib.pyplot as plt
import numpy as np
import shap
import xgboost

import weave
from weave.legacy.weave.ecosystem import huggingface as hf
from weave.legacy.weave.ecosystem import xgboost as weave_xgb


@weave.op(
    input_type={
        "df": weave.types.TypedDict({}),
    },
    output_type=weave.types.TypedDict({"X": weave.types.Any(), "y": weave.types.Any()}),
    hidden=True,
)
def split_labels(df, label_col: str):
    X = df.drop(label_col, axis=1)
    y = df[label_col]
    return {"X": X, "y": y}


@weave.type()
class ShapValues:
    values: np.ndarray

    @weave.op(
        output_type=weave.types.FileType(weave.types.Const(weave.types.String(), "png"))  # type: ignore
    )
    def summary_plot(self):
        shap.summary_plot(self.values, show=False)
        path = "/tmp/shap-%s.png" % random.randrange(0, 1000000)
        plt.savefig(path)
        plt.close()
        return weave.legacy.weave.ops.LocalFile(path)


class ShapExplanationType(weave.types.Type):
    instance_classes = shap.Explanation

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.pickle", binary=True) as f:
            pickle.dump(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.pickle", binary=True) as f:
            return pickle.load(f)


@weave.op()
def shap_explain_tree(self: xgboost.core.Booster, data: typing.Any) -> ShapValues:
    explainer = shap.TreeExplainer(self)
    shap_values = explainer.shap_values(data)
    return ShapValues(shap_values)


@weave.op()
def shap_explain(
    pipeline_output: hf.FullTextClassificationPipelineOutput,
) -> shap.Explanation:
    # TODO: shap has some options, like computing logits is better in some cases?
    # TODO: does shap work for all task styles?
    pipeline = weave.use(pipeline_output._model.pipeline())
    explainer = shap.Explainer(pipeline)
    return explainer([pipeline_output.model_input])


@weave.op()
def shap_plot_text(shap_values: shap.Explanation) -> weave.legacy.weave.ops.Html:
    html = shap.plots.text(shap_values, display=False)
    return weave.legacy.weave.ops.Html(html)


@weave.type()
class ShapPlotText(weave.Panel):
    id = "ShapPlotText"
    input_node: weave.Node[shap.Explanation]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.PanelHtml:
        return weave.legacy.weave.panels.PanelHtml(shap_plot_text(self.input_node))



"""Defines the custom Image weave type."""

from weave.trace import serializer
from weave.trace.custom_objs import MemTraceFilesArtifact

dependencies_met = False

try:
    from PIL import Image

    dependencies_met = True
except ImportError:
    pass


def save(obj: "Image.Image", artifact: MemTraceFilesArtifact, name: str) -> None:
    # Note: I am purposely ignoring the `name` here and hard-coding the filename to "image.png".
    # There is an extensive internal discussion here:
    # https://weightsandbiases.slack.com/archives/C03BSTEBD7F/p1723670081582949
    #
    # In summary, there is an outstanding design decision to be made about how to handle the
    # `name` parameter. One school of thought is that using the `name` parameter allows multiple
    # object to use the same artifact more cleanly. However, another school of thought is that
    # the payload should not be dependent on an external name - resulting in different payloads
    # for the same logical object.
    #
    # Using `image.png` is fine for now since we don't have any cases of multiple objects
    # using the same artifact. Moreover, since we package the deserialization logic with the
    # object payload, we can always change the serialization logic later without breaking
    # existing payloads.
    with artifact.new_file("image.png", binary=True) as f:
        obj.save(f, format="png")  # type: ignore


def load(artifact: MemTraceFilesArtifact, name: str) -> "Image.Image":
    # Note: I am purposely ignoring the `name` here and hard-coding the filename. See comment
    # on save.
    path = artifact.path("image.png")
    return Image.open(path)


def register() -> None:
    if dependencies_met:
        serializer.register_serializer(Image.Image, save, load)



from weave.legacy.weave.ops_arrow.arraylist_ops import *
from weave.legacy.weave.ops_arrow.boolean import *
from weave.legacy.weave.ops_arrow.convert_ops import *
from weave.legacy.weave.ops_arrow.date import *
from weave.legacy.weave.ops_arrow.dict import *
from weave.legacy.weave.ops_arrow.list_join import *
from weave.legacy.weave.ops_arrow.list_ops import *  # type: ignore
from weave.legacy.weave.ops_arrow.number import *
from weave.legacy.weave.ops_arrow.obj import *
from weave.legacy.weave.ops_arrow.ref_ops import *
from weave.legacy.weave.ops_arrow.string import *



import dataclasses
import typing

from weave.legacy.weave import weave_types as types


@dataclasses.dataclass
class Run:
    id: str
    op_name: str
    state: str = dataclasses.field(default_factory=lambda: "pending")
    prints: list[str] = dataclasses.field(default_factory=list)
    history: list[dict[str, typing.Any]] = dataclasses.field(default_factory=list)
    inputs: dict[str, typing.Any] = dataclasses.field(default_factory=dict)
    output: typing.Any = dataclasses.field(default_factory=lambda: None)

    @property
    def trace_id(self) -> str:
        raise NotImplementedError

    @property
    def ui_url(self):
        return ""


types.RunType.instance_classes = Run



import hashlib
import typing
from inspect import Parameter, signature

import pyarrow as pa

from weave.legacy.weave import weave_types
from weave.legacy.weave.api import op
from weave.legacy.weave import errors, gql_op_plugin, op_def, partial_object
from weave.legacy.weave.decorator_arrow_op import arrow_op
from weave.legacy.weave.gql_op_plugin import wb_gql_op_plugin
from weave.legacy.weave.input_provider import InputProvider
from weave.legacy.weave.ops_arrow import ArrowWeaveList, ArrowWeaveListType
from weave.legacy.weave.ops_domain import wb_domain_types

"""
This file contains utilities for constructing GQL ops (used by all the ops in
`ops_domain`). There are explicitly 4 functions:
    - `gql_prop_op`: used for getting properties of GQL objects
    - `gql_root_op`: used to start a query (much match to a query on the `Query`
      type)
    - `gql_direct_edge_op`: used to get a direct edge of a GQL object leading to
      another GQL object
        - Has an additional `is_many` arg to determine if the result is a single
          or many objects
    - `gql_connection_op`: used to get a connection of GQL objects (the standard
      edges/nodes pattern)

All but the prop_op have the ability to specify additional inputs and how to map
such inputs to a query param string.

Each of these leverage the underlying `wb_gql_op_plugin` to create an op. You
are always free to directly use `wb_gql_op_plugin` if you need to write custom
logic - these are just wrappers since these cases are so common. 

Please see `project_ops.py` for examples of all the above cases.
"""


ParamStrFn = typing.Callable[[InputProvider], str]


def _make_alias(*args: str, prefix: str = "alias") -> str:
    inputs = "_".join([str(arg) for arg in args])
    digest = hashlib.md5(inputs.encode()).hexdigest()
    return f"{prefix}_{digest}"


def _param_str(inputs: InputProvider, param_str_fn: typing.Optional[ParamStrFn]) -> str:
    param_str = ""
    if param_str_fn:
        param_str = param_str_fn(inputs)
    return param_str


def _alias(
    inputs: InputProvider,
    param_str_fn: typing.Optional[ParamStrFn],
    prop_name: str,
) -> str:
    alias = ""
    param_str = _param_str(inputs, param_str_fn)
    if param_str_fn:
        alias = f"{_make_alias(param_str, prefix=prop_name)}"
    return alias


def register_vectorized_gql_prop_op(
    scalar_op: op_def.OpDef, prop_name: str, scalar_output_type: weave_types.Type
) -> None:
    op_args = scalar_op.input_type

    first_arg_name: str
    scalar_input_type: weave_types.Type
    scalar_op_name = scalar_op.name
    first_arg_name = next(iter(op_args.to_dict()))
    original_scalar_input_type = op_args.initial_arg_types[first_arg_name]

    # require that the desired key be present on the object type
    assert isinstance(
        original_scalar_input_type, partial_object.PartialObjectTypeGeneratorType
    )
    scalar_input_type = original_scalar_input_type

    vectorized_op_name = "ArrowWeaveList" + scalar_op_name
    arrow_input_type = {first_arg_name: ArrowWeaveListType(scalar_input_type)}
    arrow_output_type = ArrowWeaveListType(scalar_output_type)

    def vectorized_gql_property_getter_op_fn(**inputs):
        self = typing.cast(ArrowWeaveList, inputs[first_arg_name])
        object_type = typing.cast(partial_object.PartialObjectType, self.object_type)
        return ArrowWeaveList(
            # keep things dictionary encoded
            pa.DictionaryArray.from_arrays(
                self._arrow_data.indices, self._arrow_data.dictionary.field(prop_name)
            ),
            object_type.keys[prop_name],
            self._artifact,
        )

    vectorized_gql_property_getter_op_fn.__signature__ = (  # type: ignore
        scalar_op.raw_resolve_fn.__signature__  # type: ignore
    )

    vectorized_gql_property_getter_op_fn.sig = scalar_op.raw_resolve_fn.sig  # type: ignore

    arrow_op(
        name=vectorized_op_name,
        input_type=arrow_input_type,
        output_type=arrow_output_type,
        plugins=scalar_op.plugins,
    )(vectorized_gql_property_getter_op_fn)


def gql_prop_op(
    op_name: str,
    input_type: weave_types.Type,
    prop_name: str,
    output_type: weave_types.Type,
    hidden: bool = False,
):
    first_arg_name = input_type.name

    def gql_property_getter_op_fn(**inputs):
        return inputs[first_arg_name][prop_name]

    sig = signature(gql_property_getter_op_fn)
    params = [Parameter(first_arg_name, Parameter.POSITIONAL_OR_KEYWORD)]
    sig = sig.replace(parameters=tuple(params))
    gql_property_getter_op_fn.__signature__ = sig  # type: ignore
    gql_property_getter_op_fn.sig = sig  # type: ignore

    scalar_op = op(
        name=op_name,
        plugins=wb_gql_op_plugin(
            lambda inputs, inner: prop_name,
        ),
        input_type={first_arg_name: input_type},
        output_type=output_type,
        hidden=hidden,
    )(gql_property_getter_op_fn)

    # side effect: register an arrow-vectorized version of the op
    register_vectorized_gql_prop_op(scalar_op, prop_name, output_type)
    return scalar_op


def _get_required_fragment(output_type: weave_types.Type):
    return (
        output_type.instance_class.REQUIRED_FRAGMENT  # type: ignore
        if output_type.instance_class is not None
        and hasattr(output_type.instance_class, "REQUIRED_FRAGMENT")
        else ""
    )


def gql_root_op(
    op_name: str,
    prop_name: str,
    output_type: weave_types.Type,
    additional_inputs_types: typing.Optional[dict[str, weave_types.Type]] = None,
    param_str_fn: typing.Optional[typing.Callable[[InputProvider], str]] = None,
):
    return gql_direct_edge_op(
        op_name,
        None,
        prop_name,
        output_type,
        additional_inputs_types,
        param_str_fn,
    )


def gql_direct_edge_op(
    op_name: str,
    input_type: typing.Optional[weave_types.Type],
    prop_name: str,
    output_type: weave_types.Type,
    additional_inputs_types: typing.Optional[dict[str, weave_types.Type]] = None,
    param_str_fn: typing.Optional[ParamStrFn] = None,
    is_many: bool = False,
):
    is_root = input_type is None
    first_arg_name: str = "gql_obj" if input_type is None else input_type.name
    if not output_type.instance_class or isinstance(
        output_type.instance_class, wb_domain_types.PartialObject
    ):
        raise ValueError(
            f"output_type must be a GQLTypeMixin, got {output_type} instead"
        )

    def query_fn(inputs, inner):
        alias = ""
        param_str = ""
        if param_str_fn is not None:
            param_str = _param_str(inputs, param_str_fn)
            alias = _alias(inputs, param_str_fn, prop_name)
            param_str = f"({param_str})"
            alias = f"{alias}:"
        return f"""
            {alias} {prop_name}{param_str} {{
                {_get_required_fragment(output_type)}
                {inner}
            }}
        """

    def key_fn(
        input_provider: InputProvider, self_type: weave_types.Type
    ) -> weave_types.Type:
        alias = _alias(input_provider, param_str_fn, prop_name)
        key = alias if alias != "" else prop_name

        ret_type = output_type
        if isinstance(self_type, partial_object.PartialObjectType):
            keys = self_type.keys
            assert isinstance(
                output_type, partial_object.PartialObjectTypeGeneratorType
            )
            new_keys = keys[key]
            if is_many:
                new_keys = typing.cast(weave_types.List, new_keys).object_type
            new_keys = typing.cast(weave_types.TypedDict, new_keys)
            ret_type = output_type.with_keys(new_keys.property_types)

        if is_many:
            ret_type = weave_types.List(ret_type)

        return ret_type

    additional_inputs_types = additional_inputs_types or {}
    if is_root:

        def gql_relationship_getter_op(**inputs):
            raise errors.WeaveGQLCompileError(
                f"{op_name} should not be executed directly. If you see this error, it is a bug in the Weave compiler."
            )

    else:

        def gql_relationship_getter_op(**inputs):
            gql_obj: partial_object.PartialObject = inputs[first_arg_name]
            additional_inputs = {
                key: value for key, value in inputs.items() if key != first_arg_name
            }
            name = prop_name
            if param_str_fn:
                param_str = param_str_fn(InputProvider(additional_inputs))
                name = _make_alias(param_str, prefix=prop_name)

            assert issubclass(output_type.instance_class, partial_object.PartialObject)

            if is_many:
                return [
                    output_type.instance_class.from_keys(item) for item in gql_obj[name]
                ]
            if (
                len(gql_obj.keys) is None
            ):  # == wb_domain_types.UntypedOpaqueDict.from_json_dict(None):
                return None
            gql_val = gql_obj[name]
            if gql_val is None:
                return None
            return output_type.instance_class.from_keys(gql_val)

    sig = signature(gql_relationship_getter_op)
    base_params = (
        [Parameter(first_arg_name, Parameter.POSITIONAL_OR_KEYWORD)]
        if not is_root
        else []
    )
    new_params = [
        *base_params,
        *[
            Parameter(key, Parameter.POSITIONAL_OR_KEYWORD)
            for key in additional_inputs_types.keys()
        ],
    ]
    sig = sig.replace(parameters=tuple(new_params))
    gql_relationship_getter_op.__signature__ = sig  # type: ignore
    gql_relationship_getter_op.sig = sig  # type: ignore

    base_input_type: dict[str, weave_types.Type] = (
        {first_arg_name: typing.cast(weave_types.Type, input_type)}
        if not is_root
        else {}
    )
    _output_type = output_type
    if is_many:
        _output_type = weave_types.List(output_type)
    gql_relationship_getter_op = op(
        name=op_name,
        plugins=wb_gql_op_plugin(query_fn, is_root, None, key_fn),
        input_type={**base_input_type, **additional_inputs_types},
        output_type=_output_type,
    )(gql_relationship_getter_op)

    return gql_relationship_getter_op


def gql_connection_op(
    op_name: str,
    input_type: weave_types.Type,
    prop_name: str,
    output_type: weave_types.Type,
    additional_inputs_types: typing.Optional[dict[str, weave_types.Type]] = None,
    param_str_fn: typing.Optional[typing.Callable[[InputProvider], str]] = None,
):
    first_arg_name = "gql_obj" if input_type is None else input_type.name
    if not output_type.instance_class or isinstance(
        output_type.instance_class, wb_domain_types.PartialObject
    ):
        raise ValueError(
            f"output_type must be a GQLTypeMixin, got {output_type} instead"
        )

    def query_fn(inputs, inner):
        alias = ""
        param_str = ""
        if param_str_fn:
            alias = _alias(inputs, param_str_fn, prop_name)
            param_str = _param_str(inputs, param_str_fn)
            param_str = f"({param_str})"
            alias = alias + ":"

        return f"""
            {alias} {prop_name}{param_str} {{
                edges {{
                    node {{
                        {_get_required_fragment(output_type)}
                        {inner}
                    }}
                }}
            }}
        """

    def key_fn(
        input_provider: InputProvider, self_type: weave_types.Type
    ) -> weave_types.Type:
        alias = _alias(input_provider, param_str_fn, prop_name)
        key = alias if alias != "" else prop_name

        if isinstance(self_type, partial_object.PartialObjectType):
            keys = self_type.keys
            assert isinstance(
                output_type, partial_object.PartialObjectTypeGeneratorType
            )
            new_keys = typing.cast(
                weave_types.TypedDict,
                typing.cast(
                    weave_types.TypedDict,
                    typing.cast(
                        weave_types.List,
                        typing.cast(weave_types.TypedDict, keys[key]).property_types[
                            "edges"
                        ],
                    ).object_type,
                ).property_types["node"],
            )
            return weave_types.List(output_type.with_keys(new_keys.property_types))

        return weave_types.List(output_type)

    additional_inputs_types = additional_inputs_types or {}

    def gql_connection_walker_op(**inputs):
        gql_obj: partial_object.PartialObject = inputs[first_arg_name]
        additional_inputs = {
            key: value for key, value in inputs.items() if key != first_arg_name
        }
        name = prop_name
        if param_str_fn:
            param_str = param_str_fn(InputProvider(additional_inputs))
            name = _make_alias(param_str, prefix=prop_name)
        # If we have a None argument, return an empty list.
        if (
            gql_obj.keys == []
        ):  #  == wb_domain_types.UntypedOpaqueDict.from_json_dict(None):
            return []
        return [
            output_type.instance_class.from_keys(edge["node"])
            for edge in gql_obj[name]["edges"]
        ]

    sig = signature(gql_connection_walker_op)
    new_params = [
        Parameter(first_arg_name, Parameter.POSITIONAL_OR_KEYWORD),
        *[
            Parameter(key, Parameter.POSITIONAL_OR_KEYWORD)
            for key in additional_inputs_types.keys()
        ],
    ]
    sig = sig.replace(parameters=tuple(new_params))
    gql_connection_walker_op.__signature__ = sig  # type: ignore
    gql_connection_walker_op.sig = sig  # type: ignore
    gql_connection_walker_op = op(
        name=op_name,
        plugins=wb_gql_op_plugin(query_fn, False, None, key_fn),
        input_type={first_arg_name: input_type, **additional_inputs_types},
        output_type=weave_types.List(output_type),
    )(gql_connection_walker_op)

    return gql_connection_walker_op


def make_root_op_gql_op_output_type(
    prop_name: str,
    param_str_fn: ParamStrFn,
    output_type: partial_object.PartialObjectTypeGeneratorType,
    use_alias: bool = False,
) -> gql_op_plugin.GQLOutputTypeFn:
    """Creates a GQLOutputTypeFn for a root op that returns a list of objects with keys."""

    def _root_op_gql_op_output_type(
        inputs: InputProvider, input_type: weave_types.Type
    ) -> weave_types.Type:
        key = (
            prop_name
            if not use_alias
            else _alias(
                inputs,
                param_str_fn,
                prop_name,
            )
        )

        key_type: weave_types.Type = typing.cast(weave_types.TypedDict, input_type)
        keys = ["instance", key, "edges", "node"]
        for key in keys:
            if isinstance(key_type, weave_types.List):
                key_type = key_type.object_type
            if isinstance(key_type, weave_types.TypedDict):
                key_type = key_type.property_types[key]

        object_type = output_type.with_keys(
            typing.cast(weave_types.TypedDict, key_type).property_types
        )
        return weave_types.List(object_type)

    return _root_op_gql_op_output_type



"""
This module provides callback support for OpenAI's Chat API.  Users can define custom behaviour at various stages of the API call lifecycle (e.g. after yield streaming chunk, before return)

We provide simple patch and unpatch functions to enable or disable these features.

By default, the patch method will provide goodies like auto-collecting streaming completions and logging results to StreamTable
"""

__all__ = ["patch", "unpatch"]

from .openai import patch, unpatch



import datetime
import hashlib
import itertools
import string

import pyarrow as pa
import pytest
from PIL import Image

from weave.legacy.weave import api as weave
from weave.legacy.weave import (
    box,
    context_state,
    errors,
    graph,
    mappers_arrow,
    ops,
    storage,
    weave_internal,
)

# If you're thinking of import vectorize here, don't! Put your
# tests in test_arrow_vectorizer.py instead
from weave.legacy.weave import ops_arrow as arrow
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.arrow import constructors
from weave.legacy.weave.arrow.arrow_tags import (
    recursively_encode_pyarrow_strings_as_dictionaries,
)
from weave.legacy.weave.language_features.tagging import (
    make_tag_getter_op,
    tag_store,
    tagged_value_type,
)
from weave.legacy.weave.op_def import map_type
from weave.legacy.weave.ops_domain import project_ops
from weave.legacy.weave.ops_primitives import list_, make_list
from .util import list_arrow_test_helpers as lath

from weave.legacy.tests.util import tag_test_util as ttu
from weave.legacy.tests.util import weavejs_ops
from . import test_wb

_loading_builtins_token = context_state.set_loading_built_ins()
# T in `conftest::pre_post_each_test` we set a custom artifact directory for each test for isolation
# Puting this import in the context allows the test execution to access these ops


@weave.type()
class Point2:
    x: float
    y: float

    @weave.op()
    def get_x(self) -> float:
        return self.x


context_state.clear_loading_built_ins(_loading_builtins_token)


def simple_hash(n, b):
    return int.from_bytes(hashlib.sha256(str(n).encode()).digest(), "little") % b


def create_arrow_data(n_rows, n_extra_cols=0, images=False):
    inner_count = int(n_rows / 25)
    base_im = Image.linear_gradient("L")
    x_choices = string.ascii_lowercase
    extra_cols = [chr(ord("a") + i) for i in range(n_extra_cols)]
    ims = []
    for i, (rotate, shear, _) in enumerate(
        itertools.product(range(5), range(5), range(inner_count))
    ):
        im = {
            "rotate": rotate,
            "shear": shear,
            "x": x_choices[simple_hash(i**13, 3)],
            "y": simple_hash(i, 10),
        }
        if images:
            im["image"] = base_im.rotate(rotate * 4).transform(
                (256, 256), Image.AFFINE, (1, shear / 10, 0, 0, 1, 0), Image.BICUBIC
            )
        for j, col in enumerate(extra_cols):
            im[col] = x_choices[simple_hash(i * 13**j, 11)]
        ims.append(im)
    arr = arrow.to_arrow(ims)
    return storage.save(arr)


def test_groupby_index_count():
    ref = create_arrow_data(1000)
    node = (
        weave.get(ref)
        .groupby(lambda row: ops.dict_(rotate=row["rotate"], shear=row["shear"]))[1]
        .count()
    )
    assert weave.use(node) == 40


@pytest.mark.parametrize(
    "sort_lambda, sort_dirs, exp_rotation_avg",
    [
        (lambda row: list_.make_list(a=row.groupkey().pick("rotate")), ["asc"], 0),
        (lambda row: list_.make_list(a=row.groupkey().pick("rotate")), ["desc"], 4),
        # This does not work yet since vectorized groupby does not know how to pick
        # (or many ops for that matter. The only sort operations that work right now
        # are those on the key itself. Luckily, that is the only thing you can construct
        # in the UI anyway)
        # (lambda row: list_.make_list(a=row.pick("rotate").avg()), ["desc"], 4),
        (
            lambda row: list_.make_list(
                a=row.groupkey().pick("shear"), b=row.groupkey().pick("rotate")
            ),
            ["asc", "desc"],
            4,
        ),
    ],
)
def test_groupby_sort(sort_lambda, sort_dirs, exp_rotation_avg):
    ref = create_arrow_data(1000)
    grouped_node = weave.get(ref).groupby(
        lambda row: ops.dict_(rotate=row["rotate"], shear=row["shear"])
    )
    sorted_node = grouped_node.sort(sort_lambda, sort_dirs)
    first_group = sorted_node[0]
    first_group_rotations = first_group.pick("rotate")
    first_group_rotation_avg = first_group_rotations.avg()
    assert weave.use(first_group_rotation_avg) == exp_rotation_avg


def test_js_groupby_sort():
    list_data = [{"a": 1, "b": 1}, {"a": 1, "b": 2}, {"a": 2, "b": 1}, {"a": 2, "b": 2}]
    list_node = list_.make_list(**{f"{n}": v for n, v in enumerate(list_data)})
    arrow_node = weave.save(arrow.to_arrow(list_data))
    node = weavejs_ops.groupby(
        list_node,
        weave_internal.define_fn(
            {"row": list_node.type.object_type},
            lambda row: ops.dict_(
                a=graph.OutputNode(
                    types.String(),
                    "pick",
                    {"obj": row, "key": graph.ConstNode(types.String(), "a")},
                )
            ),
        ),
    )
    # Critical replacement of the input to use arrow!
    node.from_op.inputs["arr"] = arrow_node
    node = weavejs_ops.sort(
        node,
        weave_internal.define_fn(
            {"row": node.type.object_type},
            lambda row: ops.make_list(
                a=graph.OutputNode(
                    types.String(),
                    "pick",
                    {
                        "obj": graph.OutputNode(
                            types.TypedDict({"a": types.String()}),
                            "group-groupkey",
                            {"obj": row},
                        ),
                        "key": graph.ConstNode(types.String(), "a"),
                    },
                )
            ),
        ),
        ops.make_list(a="asc"),
    )
    assert weave.use(node) != None


def test_group_key():
    data = weave.save(arrow.to_arrow([1, 2, 3]))
    res = data.groupby(lambda row: row + 1)[0].groupkey()
    assert weave.use(res) == 2


def test_map_scalar_map():
    ref = create_arrow_data(100)

    node = weave.get(ref).map(lambda row: row["y"] + 1).map(lambda row: row + 9)
    assert weave.use(node[0]) == 15
    assert weave.use(node[4]) == 17


def test_groupby_mapped_groupby():
    ref = create_arrow_data(1000)
    node = (
        weave.get(ref)
        .groupby(lambda row: ops.dict_(rotate=row["rotate"], shear=row["shear"]))
        .map(lambda row: row.groupby(lambda row: row["y"]))
        .dropna()
        .count()
    )
    assert weave.use(node) == 25


def test_table_string_histogram():
    ref = create_arrow_data(1000)
    node = (
        weave.get(ref)
        .groupby(lambda row: ops.dict_(rotate=row["rotate"]))[0]
        .pick("x")
        .groupby(lambda row: ops.dict_(row=row))
        .map(lambda row: row.groupkey().merge(ops.dict_(count=row.count())))
        .count()
    )
    assert weave.use(node) == 3


def test_custom_types():
    data_node = arrow.to_weave_arrow(
        [
            {"a": 5, "im": Image.linear_gradient("L").rotate(0)},
            {"a": 6, "im": Image.linear_gradient("L").rotate(4)},
        ]
    )
    assert weave.use(data_node[0]["im"].width_()) == 256

    assert weave.use(data_node.map(lambda row: row["im"].width_())).to_pylist_raw() == [
        256,
        256,
    ]


def test_custom_types_tagged():
    im1 = tag_store.add_tags(Image.linear_gradient("L").rotate(0), {"a": 1})
    im2 = tag_store.add_tags(Image.linear_gradient("L").rotate(4), {"a": 2})

    data_node = weave.save(arrow.to_arrow([{"a": 5, "im": im1}, {"a": 6, "im": im2}]))
    width_node = data_node[0]["im"].width_()
    assert weave.use(width_node) == 256

    assert width_node.type == tagged_value_type.TaggedValueType(
        types.TypedDict({"a": types.Int()}), types.Int()
    )

    mapped_width_node = data_node.map(lambda row: row["im"].width_())
    assert weave.use(mapped_width_node).to_pylist_raw() == [
        256,
        256,
    ]

    assert mapped_width_node.type == arrow.ArrowWeaveListType(
        tagged_value_type.TaggedValueType(
            types.TypedDict({"a": types.Int()}), types.Int()
        )
    )


def test_custom_saveload():
    data = arrow.to_arrow(
        [
            {"a": 5, "im": Image.linear_gradient("L").rotate(0)},
            {"a": 6, "im": Image.linear_gradient("L").rotate(4)},
        ]
    )
    ref = storage.save(data)
    data2 = storage.get(str(ref))
    # print("data2", data2._artifact)
    assert weave.use(weave_internal.const(data2)[0]["im"].width_()) == 256


def test_custom_in_list_saveload():
    data = arrow.to_arrow(
        [
            {"a": 5, "im": [Image.linear_gradient("L").rotate(0)]},
            {"a": 6, "im": [Image.linear_gradient("L").rotate(4)]},
        ]
    )
    ref = storage.save(data)
    data2 = storage.get(str(ref))
    # print("data2", data2._artifact)
    assert weave.use(weave_internal.const(data2)[0]["im"].width_()) == [256]


def test_custom_tagged_groupby1():
    im1 = tag_store.add_tags(Image.linear_gradient("L").rotate(0), {"a": 1})
    im2 = tag_store.add_tags(Image.linear_gradient("L").rotate(4), {"a": 2})
    raw_data = box.box([{"a": 5, "im": im1}, {"a": 6, "im": im2}])
    data_node = weave.save(arrow.to_arrow(raw_data))
    grouped_node = data_node.groupby(lambda row: ops.dict_(a=row["a"]))
    group1_node = grouped_node[0]

    assert grouped_node.type == arrow.ops.awl_group_by_result_type(
        types.TypedDict(
            {
                "a": types.Int(),
                "im": tagged_value_type.TaggedValueType(
                    types.TypedDict({"a": types.Int()}),
                    weave.type_of(Image.linear_gradient("L").rotate(4)),
                ),
            }
        ),
        types.TypedDict({"a": types.Int()}),
    )

    assert group1_node.type == arrow.ops.awl_group_by_result_object_type(
        types.TypedDict(
            {
                "a": types.Int(),
                "im": tagged_value_type.TaggedValueType(
                    types.TypedDict({"a": types.Int()}),
                    weave.type_of(Image.linear_gradient("L").rotate(4)),
                ),
            }
        ),
        types.TypedDict({"a": types.Int()}),
    )

    assert (
        weave.use(
            data_node.groupby(lambda row: ops.dict_(a=row["a"]))[0]
            .pick("im")
            .offset(0)[0]
            .width_()
        )
        == 256
    )


def test_custom_tagged_groupby2():
    im1 = tag_store.add_tags(Image.linear_gradient("L").rotate(0), {"a": 1})
    im2 = tag_store.add_tags(Image.linear_gradient("L").rotate(4), {"a": 2})
    raw_data = box.box([{"a": 5, "im": im1}, {"a": 6, "im": im2}])

    tag_store.add_tags(raw_data, {"list_tag": 3})

    data_node = weave.save(arrow.to_arrow(raw_data))
    grouped_node = data_node.groupby(lambda row: ops.dict_(inner_group_key=row["a"]))
    target = tagged_value_type.TaggedValueType(
        types.TypedDict(
            {
                "list_tag": types.Int(),
            }
        ),
        arrow.ops.awl_group_by_result_type(
            types.TypedDict(
                {
                    "a": types.Int(),
                    "im": tagged_value_type.TaggedValueType(
                        types.TypedDict({"a": types.Int()}),
                        weave.type_of(Image.linear_gradient("L").rotate(4)),
                    ),
                }
            ),
            # types.TypedDict({"a": types.Int()}),
            types.TypedDict(
                {
                    "inner_group_key": tagged_value_type.TaggedValueType(
                        types.TypedDict({"list_tag": types.Int()}), types.Int()
                    )
                }
            ),
        ),
    )
    assert grouped_node.type == target

    get_list_tag = make_tag_getter_op.make_tag_getter_op("list_tag", types.Int())
    assert weave.use(get_list_tag(grouped_node)) == 3


def test_custom_groupby_1():
    data = arrow.to_weave_arrow(
        [
            {"a": 5, "im": Image.linear_gradient("L").rotate(0)},
            {"a": 6, "im": Image.linear_gradient("L").rotate(4)},
        ]
    )

    assert (
        weave.use(
            data.groupby(lambda row: ops.dict_(a=row["a"]))[0]
            .pick("im")
            .offset(0)[0]
            .width_()
        )
        == 256
    )


def test_custom_groupby_intermediate_save():
    data = arrow.to_weave_arrow(
        [
            {"a": 5, "im": Image.linear_gradient("L").rotate(0)},
            {"a": 6, "im": Image.linear_gradient("L").rotate(4)},
        ]
    )
    node = data.groupby(lambda row: ops.dict_(a=row["a"]))[0]
    saved_node = weave.save(node, "test_custom_groupby_intermediate_save:latest")
    weave.use(saved_node)
    loaded_node = ops.get(
        "local-artifact:///test_custom_groupby_intermediate_save:latest/obj"
    )
    assert weave.use(loaded_node.pick("im").offset(0)[0].width_()) == 256


def test_map_array():
    data = arrow.to_weave_arrow([1, 2, 3])
    assert weave.use(data.map(lambda i: i + 1)).to_pylist_raw() == [2, 3, 4]


def test_map_typeddict():
    data = arrow.to_weave_arrow([{"a": 1, "b": 2}, {"a": 3, "b": 5}])
    assert weave.use(data.map(lambda row: row["a"])).to_pylist_raw() == [1, 3]


@pytest.mark.parametrize(
    "data",
    [
        # ELEMENT TYPE:
        ## 0 Depth
        ### Primitive
        [1, 2],
        ## 1 Depth
        ### list
        [[1], [2]],
        ### dict
        [{"outer_a": 1}, {"outer_a": 2}],
        ## 2 Depth
        ### dict of list
        [{"outer_a": [1]}, {"outer_a": [2]}],
        ### dict of dict
        [{"outer_a": {"inner_a": 1}}, {"outer_a": {"inner_a": 2}}],
        ### list of list
        # [[[1]], [[2]]],
        ### list of dict
        [[{"outer_a": 1}], [{"outer_a": 1}]],
        ## 3 Depth
        ### dict of list of list
        [{"outer_a": [[1]]}, {"outer_a": [[2]]}],
        ### dict of list of dict
        [{"outer_a": [{"inner_a": 1}]}, {"outer_a": [{"inner_a": 2}]}],
        ### dict of dict of list
        [{"outer_a": {"inner_a": [1]}}, {"outer_a": {"inner_a": [2]}}],
        ### dict of dict of dict
        [
            {"outer_a": {"inner_a": {"inner_inner_a": 1}}},
            {"outer_a": {"inner_a": {"inner_inner_a": 2}}},
        ],
        ### list of list of list
        [[[[1]]], [[[2]]]],
        ### list of list of dict
        [[[{"outer_a": 1}]], [[{"outer_a": 2}]]],
        ### list of dict of list
        [[{"outer_a": [1]}], [{"outer_a": [2]}]],
        ### list of dict of dict
        [[{"outer_a": {"inner_a": 1}}], [{"outer_a": {"inner_a": 2}}]],
    ],
)
def test_arrow_nested_identity(data):
    assert weave.use(weave.save(arrow.to_weave_arrow(data))[0]) == data[0]


def test_arrow_nested_with_refs():
    # As of this writing, the first time we call `to_arrow`, the underlying
    # `save` operation does not convert relative artifact paths to URIs. This is
    # in `ArrowWeaveList::save_instance`.
    data_node = arrow.to_arrow(
        [{"outer": [{"inner": Image.linear_gradient("L").rotate(0)}]}]
    )

    raw_path = data_node._arrow_data[0]["outer"][0]["inner"].as_py()

    raw_data = weave.use(weave.save(data_node))
    img_entry_data = raw_data._arrow_data[0]["outer"][0]["inner"].as_py()

    # Next, we get a derive node from the data_node, and assert that the path is
    # converted to an artifact reference when appropriate.
    col_node = arrow.ops.pick(weave_internal.const(data_node), "outer")
    # Note: we don't need to save the `col_node` because
    # they are already converted to the node representation via dispatch
    raw_col_data = weave.use(col_node)
    img_entry_col = raw_col_data._arrow_data[0][0]["inner"].as_py()
    # For this case we want the entry to be the short-form - since the underlying artifact is the same
    assert raw_path == img_entry_col
    assert data_node._artifact == raw_col_data._artifact

    # Finally, when we have a new artifact (we can force this by saving)
    # the path is converted to an artifact reference
    raw_col_data = weave.use(weave.save(col_node, "NewArtifact"))
    img_entry_col = raw_col_data._arrow_data[0][0]["inner"].as_py()
    assert img_entry_data == img_entry_col
    assert data_node._artifact != raw_col_data._artifact


def test_map_object():
    data = arrow.to_weave_arrow([Point2(1, 2), Point2(5, 6)])
    assert weave.use(data.map(lambda row: row.get_x())).to_pylist_raw() == [1, 5]


@pytest.mark.skip("not working yet")
def test_map_typeddict_object():
    data = arrow.to_weave_arrow(
        [{"a": 0, "p": Point2(1, 2)}, {"a": 3, "p": Point2(9, 12)}]
    )
    assert weave.use(data.map(lambda row: row["p"])).to_pylist_raw() == []


def test_arrow_list_of_ref_to_item_in_list():
    l = [{"a": 5, "b": 6}, {"a": 7, "b": 9}]
    l_node = weave.save(l, "my-l")

    list_dict_with_ref = arrow.to_arrow([{"c": l_node[0]["a"]}, {"c": l_node[1]["a"]}])
    d_node = weave.save(list_dict_with_ref, "my-dict_with_ref")

    assert weave.use(d_node[0]["c"] == 5) == True
    assert weave.use(d_node[1]["c"] == 7) == True


def test_arrow_tag_propagation():
    list = box.box([1, 2, 3])
    awl = arrow.to_arrow(list)
    tag_store.add_tags(awl, {"mytag": "test"})

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("mytag", types.String())

    awl_node = weave.save(awl)
    element_node = awl_node[0]
    assert isinstance(element_node.type, tagged_value_type.TaggedValueType)
    tag_value = weave.use(tag_getter_op(element_node))
    assert tag_value == "test"


def test_arrow_element_tagging():
    list = [1, 2, 3]
    for i, elem in enumerate(list):
        taggable = box.box(elem)
        list[i] = tag_store.add_tags(taggable, {"mytag": f"test{elem}"})

    awl = arrow.to_weave_arrow(list)
    tag_getter_op = make_tag_getter_op.make_tag_getter_op("mytag", types.String())

    awl_node = weave.save(awl)
    element_node = awl_node[0]
    assert isinstance(element_node.type, tagged_value_type.TaggedValueType)
    tag_value = weave.use(tag_getter_op(element_node))
    assert tag_value == "test1"


def test_nested_arrow_element_tagging():
    list = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
    for i, elem in enumerate(list):
        taggable = box.box(elem["a"])
        list[i]["a"] = tag_store.add_tags(taggable, {"mytag": f"test{elem['a'] + 1}"})

    awl = arrow.to_arrow(list)
    tag_getter_op = make_tag_getter_op.make_tag_getter_op("mytag", types.String())
    awl_node = weave.save(awl)
    element_node = awl_node[0]
    assert isinstance(
        element_node.type.property_types["a"], tagged_value_type.TaggedValueType
    )
    tag_value = weave.use(tag_getter_op(element_node["a"]))
    assert tag_value == "test2"
    assert weave.use(element_node) == list[0]


def test_tagging_concat():
    awls = []
    for index in range(2):
        list = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        for i, elem in enumerate(list):
            taggable = box.box(elem["a"])
            list[i]["a"] = tag_store.add_tags(
                taggable, {"element_tag": f"test{elem['a'] + 1}"}
            )

        awl = arrow.to_arrow(list)
        tag_store.add_tags(awl, {"list_tag": f"index{index}"})
        awls.append(weave.save(awl))

    list_nodes = make_list(l1=awls[0], l2=awls[1])
    concatenated = arrow.ops.concat(list_nodes)

    assert weave.use(concatenated.to_py()) == [{"a": 1, "b": 2}, {"a": 3, "b": 4}] * 2
    assert concatenated.type == arrow.ArrowWeaveListType(
        tagged_value_type.TaggedValueType(
            types.TypedDict({"list_tag": types.String()}),
            types.TypedDict(
                {
                    "a": tagged_value_type.TaggedValueType(
                        types.TypedDict({"element_tag": types.String()}),
                        types.Int(),
                    ),
                    "b": types.Int(),
                }
            ),
        )
    )

    get_element_tag = make_tag_getter_op.make_tag_getter_op(
        "element_tag", types.String()
    )

    assert weave.use(get_element_tag(concatenated[0]["a"])) == "test2"

    # test tag merging
    assert concatenated[0]["a"].type == tagged_value_type.TaggedValueType(
        types.TypedDict({"element_tag": types.String(), "list_tag": types.String()}),
        types.Int(),
    )

    get_list_tag = make_tag_getter_op.make_tag_getter_op("list_tag", types.String())
    assert weave.use(get_list_tag(concatenated[0]["a"])) == "index0"


def test_arrow_unnest():
    data = arrow.to_arrow([{"a": [1, 2, 3], "b": "c"}, {"a": [4, 5, 6], "b": "d"}])
    assert weave.type_of(data) == arrow.ArrowWeaveListType(
        types.TypedDict({"a": types.List(types.Int()), "b": types.String()})
    )
    unnest_node = weave_internal.const(data).unnest()
    assert unnest_node.type == arrow.ArrowWeaveListType(
        types.TypedDict({"a": types.Int(), "b": types.String()})
    )
    assert weave.use(weave_internal.const(data).unnest()).to_pylist_raw() == [
        {"a": 1, "b": "c"},
        {"a": 2, "b": "c"},
        {"a": 3, "b": "c"},
        {"a": 4, "b": "d"},
        {"a": 5, "b": "d"},
        {"a": 6, "b": "d"},
    ]


def test_arrow_unnest_two_list_cols():
    data = arrow.to_arrow(
        [
            {"a": [1, 2, 3], "b": "c", "c": ["a", "b", "c"]},
            {"a": [4, 5, 6], "b": "d", "c": ["d", "e", "f"]},
        ]
    )
    assert weave.type_of(data) == arrow.ArrowWeaveListType(
        types.TypedDict(
            {
                "a": types.List(types.Int()),
                "b": types.String(),
                "c": types.List(types.String()),
            }
        )
    )
    unnest_node = weave_internal.const(data).unnest()
    assert unnest_node.type == arrow.ArrowWeaveListType(
        types.TypedDict({"a": types.Int(), "b": types.String(), "c": types.String()})
    )
    assert weave.use(weave_internal.const(data).unnest()).to_pylist_raw() == [
        {"a": 1, "b": "c", "c": "a"},
        {"a": 2, "b": "c", "c": "b"},
        {"a": 3, "b": "c", "c": "c"},
        {"a": 4, "b": "d", "c": "d"},
        {"a": 5, "b": "d", "c": "e"},
        {"a": 6, "b": "d", "c": "f"},
    ]


def test_arrow_nullable_concat():
    ca1 = pa.chunked_array([[1, 2], [3, 4]])
    ca2 = pa.compute.add(ca1, 1)
    awl1 = arrow.ArrowWeaveList(ca1)
    awl2 = arrow.ArrowWeaveList(ca2)
    list_of_awl = make_list(A=awl1, B=awl2, C=weave.save(None))
    result = list_of_awl.concat()
    assert weave.use(result).to_pylist_raw() == [1, 2, 3, 4, 2, 3, 4, 5]

    # Second pass - forcing none type to be first in member list
    list_of_awl.type = types.List(
        types.union(types.NoneType(), arrow.ArrowWeaveListType(types.Int()))
    )
    result = list_of_awl.concat()
    assert weave.use(result).to_pylist_raw() == [1, 2, 3, 4, 2, 3, 4, 5]


def test_arrow_weave_list_groupby_struct_type_table():
    table = pa.Table.from_pylist(
        [{"d": {"a": 1, "b": 2}, "c": 1}, {"d": {"a": 3, "b": 4}, "c": 2}]
    )
    awl = arrow.ArrowWeaveList(table)

    def group_func_body(row):
        return ops.dict_(**{"d": row["d"]})

    # note: access more than 1 level deep is broken, should work, fix
    # def group_func_body(row):
    #    return ops.dict_(**{"d.a": row["d"]["a"]})

    group_func = weave_internal.define_fn({"row": awl.object_type}, group_func_body)
    grouped = weave_internal.const(awl).groupby(group_func)
    assert weave.use(grouped[0]).to_pylist_raw() == [{"d": {"a": 1, "b": 2}, "c": 1}]


def test_arrow_weave_list_groupby_struct_chunked_array_type():
    ref = create_arrow_data(1000)

    # chunkedarray is used when there are two levels of nesting
    node = (
        weave.get(ref)
        .groupby(lambda row: ops.dict_(rotate=row["rotate"], shear=row["shear"]))
        .map(lambda row: row.groupby(lambda row: row["y"]))
        .dropna()[0][0]
    )

    assert (
        weave.use(node).to_pylist_raw()
        == [{"rotate": 0, "shear": 0, "x": "a", "y": 5}] * 5
    )


def _make_tagged_awl():
    to_tag = box.box(["a", "b", "c"])
    for i, elem in enumerate(to_tag):
        taggable = box.box(elem)
        to_tag[i] = tag_store.add_tags(taggable, {"a": f"{elem}"})
    to_tag = tag_store.add_tags(to_tag, {"outer": "tag"})
    return weave.save(arrow.to_arrow(to_tag))


def test_arrow_dict_map_tagged():
    tagged_awl = _make_tagged_awl()
    expected_output = [{"a": "a", "b": 1}, {"a": "b", "b": 1}, {"a": "c", "b": 1}]
    weave_func = lambda row: ops.dict_(a=row, b=1)
    fn = weave_internal.define_fn(
        {"row": tagged_awl.type.object_type},
        weave_func,
    )

    called = tagged_awl.map(fn)
    awl = weave.use(called)
    assert awl.to_pylist_notags() == expected_output

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("outer", types.String())
    tag_node = tag_getter_op(called[0]["a"])
    assert weave.use(tag_node) == "tag"

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("a", types.String())
    tag_node = tag_getter_op(called[0]["a"])
    assert weave.use(tag_node) == "a"


def test_arrow_filter_tagged():
    tagged_awl = _make_tagged_awl()
    expected_output = ["b", "c"]
    weave_func = lambda row: (row != "a")
    fn = weave_internal.define_fn(
        {"row": tagged_awl.type.object_type},
        weave_func,
    )

    called = tagged_awl.filter(fn)
    awl = weave.use(called)
    assert awl.to_pylist_notags() == expected_output

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("outer", types.String())
    tag_node = tag_getter_op(called[0])
    assert weave.use(tag_node) == "tag"

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("a", types.String())
    tag_node = tag_getter_op(called[0])
    assert weave.use(tag_node) == "b"


def test_arrow_sort_tagged():
    tagged_awl = _make_tagged_awl()
    expected_output = ["c", "b", "a"]
    weave_func = lambda row: list_.make_list(a=row)
    fn = weave_internal.define_fn(
        {"row": tagged_awl.type.object_type},
        weave_func,
    )

    called = tagged_awl.sort(fn, ["desc"])
    awl = weave.use(called)
    assert awl.to_pylist_notags() == expected_output

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("outer", types.String())
    tag_node = tag_getter_op(called[0])
    assert weave.use(tag_node) == "tag"

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("a", types.String())
    tag_node = tag_getter_op(called[0])
    assert weave.use(tag_node) == "c"


def test_arrow_filter_nulls():
    awl = weave.save(arrow.to_weave_arrow([-1, 0, 1, None]))
    weave_func = lambda row: row < 1
    fn = weave_internal.define_fn(
        {"row": awl.type.object_type},
        weave_func,
    )

    called = awl.filter(fn)
    awl = weave.use(called)
    assert awl.to_pylist_notags() == [-1, 0]


def test_grouped_typed_dict_assign():
    assert types.List(types.TypedDict(property_types={})).assign_type(
        arrow.ops.awl_group_by_result_object_type(
            object_type=types.TypedDict(
                property_types={"a": types.Int(), "im": types.Int()}
            ),
            _key=types.TypedDict(property_types={"a": types.String()}),
        )
    )


def test_arrow_index_var():
    data = arrow.to_weave_arrow([1, 2, 3])
    result = data.map(lambda row, index: row + index)
    assert weave.use(result).to_pylist_raw() == [1, 3, 5]


def test_concat_multiple_table_types():
    datal = weave.save(
        arrow.to_weave_arrow([{"prompt": "a"}, {"prompt": None}, {"prompt": "b"}])
    )
    datar = weave.save(
        arrow.to_weave_arrow(
            [
                {"prompt": None, "generation_prompt": "a"},
                {"prompt": "d", "generation_prompt": None},
                {"prompt": "e", "generation_prompt": "f"},
            ]
        )
    )

    to_concat = ops.make_list(l=datal, r=datar)
    result = arrow.ops.concat(to_concat)

    assert result.type == arrow.ArrowWeaveListType(
        object_type=types.TypedDict(
            property_types={
                "prompt": types.UnionType(types.String(), types.NoneType()),
                "generation_prompt": types.UnionType(types.String(), types.NoneType()),
            }
        )
    )

    result = weave.use(result)

    assert result.to_pylist_notags() == [
        {"prompt": "a", "generation_prompt": None},
        {"prompt": None, "generation_prompt": None},
        {"prompt": "b", "generation_prompt": None},
        {"prompt": None, "generation_prompt": "a"},
        {"prompt": "d", "generation_prompt": None},
        {"prompt": "e", "generation_prompt": "f"},
    ]


def test_concat_multiple_table_types_tagged():
    raw_datal = [{"prompt": "a"}, {"prompt": None}, {"prompt": "b"}]
    raw_datar = [
        {"prompt": None, "generation_prompt": "a"},
        {"prompt": "d", "generation_prompt": None},
        {"prompt": "e", "generation_prompt": "f"},
    ]

    tagged_datal = tag_store.add_tags(box.box(raw_datal), {"tag": "datal"})
    tagged_datar = tag_store.add_tags(box.box(raw_datar), {"tag": "datar"})

    datal = weave.save(arrow.to_arrow(tagged_datal))
    datar = weave.save(arrow.to_arrow(tagged_datar))

    to_concat = ops.make_list(l=datal, r=datar)
    result = arrow.ops.concat(to_concat)

    expected = arrow.ArrowWeaveListType(
        object_type=tagged_value_type.TaggedValueType(
            types.TypedDict({"tag": types.String()}),
            types.TypedDict(
                property_types={
                    "prompt": types.UnionType(types.String(), types.NoneType()),
                    "generation_prompt": types.UnionType(
                        types.String(), types.NoneType()
                    ),
                }
            ),
        )
    )

    assert result.type == expected

    result = weave.use(result)
    assert result.to_pylist_notags() == [
        {"prompt": "a", "generation_prompt": None},
        {"prompt": None, "generation_prompt": None},
        {"prompt": "b", "generation_prompt": None},
        {"prompt": None, "generation_prompt": "a"},
        {"prompt": "d", "generation_prompt": None},
        {"prompt": "e", "generation_prompt": "f"},
    ]


def test_arrow_union_int_string():
    data = [1, "a", 2]
    result = weave.save(arrow.to_arrow(data))
    assert result.type == arrow.ArrowWeaveListType(
        object_type=types.UnionType(types.Int(), types.String())
    )

    assert weave.use(result).to_pylist_raw() == data


def test_arrow_union_int_string_custom_type():
    data = [1, "a", 2, Point2(1, 2)]
    result = weave.save(arrow.to_arrow(data))
    assert result.type == arrow.ArrowWeaveListType(
        object_type=types.UnionType(types.Int(), types.String(), Point2.WeaveType())
    )

    awl = weave.use(result)
    assert [awl._index(i) for i in range(len(data))] == data


def test_arrow_union_two_struct_types_as_single_struct_type():
    data = [{"a": 1}, {"b": 2, "a": 1}, {"b": 3}]
    result = weave.save(arrow.to_arrow(data))
    assert result.type == arrow.ArrowWeaveListType(
        object_type=types.TypedDict(
            {
                "a": types.UnionType(types.Int(), types.NoneType()),
                "b": types.UnionType(types.Int(), types.NoneType()),
            }
        )
    )

    awl = weave.use(result)
    assert awl.to_pylist_notags() == [
        {"a": 1, "b": None},
        {"b": 2, "a": 1},
        {"b": 3, "a": None},
    ]


def test_arrow_concat_nested_union():
    raw_datal = [{"b": 1}]
    raw_datar = [{"b": "c"}]

    datal = weave.save(arrow.to_arrow(raw_datal))
    datar = weave.save(arrow.to_arrow(raw_datar))

    to_concat = ops.make_list(l=datal, r=datar)
    result = arrow.ops.concat(to_concat)

    expected = arrow.ArrowWeaveListType(
        object_type=types.TypedDict(
            property_types={
                "b": types.UnionType(types.String(), types.Int()),
            }
        ),
    )

    assert result.type == expected

    result = weave.use(result)
    assert result.to_pylist_notags() == [
        {"b": 1},
        {"b": "c"},
    ]


def test_arrow_concat_nested_union_with_optional_type():
    raw_datal = [{"b": 1}]
    raw_datar = [{"b": "c", "c": 3.3}]

    datal = weave.save(arrow.to_arrow(raw_datal))
    datar = weave.save(arrow.to_arrow(raw_datar))

    to_concat = ops.make_list(l=datal, r=datar)
    result = arrow.ops.concat(to_concat)

    expected = arrow.ArrowWeaveListType(
        object_type=types.TypedDict(
            property_types={
                "b": types.UnionType(types.String(), types.Int()),
                "c": types.UnionType(types.Float(), types.NoneType()),
            }
        ),
    )

    assert result.type == expected

    result = weave.use(result)
    assert result.to_pylist_notags() == [
        {"b": 1, "c": None},
        {"b": "c", "c": 3.3},
    ]


def test_arrow_concat_nested_union_with_optional_type_and_custom_type():
    raw_datal = [{"b": 1, "x": Point2(1, 2)}]
    raw_datar = [{"b": "c", "c": 3.3}]

    datal = weave.save(arrow.to_arrow(raw_datal))
    datar = weave.save(arrow.to_arrow(raw_datar))

    to_concat = ops.make_list(l=datal, r=datar)
    result = arrow.ops.concat(to_concat)

    expected = arrow.ArrowWeaveListType(
        object_type=types.TypedDict(
            property_types={
                "b": types.UnionType(types.String(), types.Int()),
                "c": types.UnionType(types.Float(), types.NoneType()),
                "x": types.UnionType(Point2.WeaveType(), types.NoneType()),
            }
        ),
    )

    assert result.type == expected

    result = weave.use(result)
    assert [result._index(i) for i in range(len(result))] == [
        {"b": 1, "c": None, "x": Point2(1, 2)},
        {"b": "c", "c": 3.3, "x": None},
    ]


def test_arrow_concat_degenerate_types():
    raw_datal = [{"b": 1, "x": Point2(1, 2)}]
    raw_datar = [{"b": "c", "x": {"x": 1, "y": 2}}]

    datal = weave.save(arrow.to_arrow(raw_datal))
    datar = weave.save(arrow.to_arrow(raw_datar))

    to_concat = ops.make_list(l=datal, r=datar)
    result = arrow.ops.concat(to_concat)

    expected = arrow.ArrowWeaveListType(
        object_type=types.TypedDict(
            property_types={
                "b": types.UnionType(types.String(), types.Int()),
                "x": types.UnionType(
                    Point2.WeaveType(),
                    types.TypedDict({"x": types.Int(), "y": types.Int()}),
                ),
            }
        ),
    )

    assert result.type == expected

    result = weave.use(result)
    assert [result._index(i) for i in range(len(result))] == [
        {"b": 1, "x": Point2(1, 2)},
        {"b": "c", "x": {"x": 1, "y": 2}},
    ]


@pytest.mark.parametrize("li", lath.ListInterfaces)
def test_arrow_timestamp_conversion(li):
    dates = [
        # Ensure these two are actually different date times! The timezone is utc in
        # our ci environment, so serializing these results in the same thing if the
        # timestamps match. We use the serialized representation for graph deduplication,
        # meaning we end up with the same node for both of these, which breaks the test
        # (but not the actual behavior of the code).
        datetime.datetime(2020, 1, 2, 3, 4, 6),
        datetime.datetime(2020, 1, 2, 3, 4, 5, tzinfo=datetime.timezone.utc),
    ]
    utc_dates = [d.astimezone(datetime.timezone.utc) for d in dates]
    timestamps = [d.timestamp() * 1000 for d in utc_dates]

    # Direct datetime type
    data = li.make_node(dates)
    if li == lath.ArrowNode:
        # Arrow converts to UTC on read out
        assert li.use_node(data) == utc_dates
    else:
        assert li.use_node(data) == dates

    # Basic Floats to datetime conversion
    data = li.make_node(timestamps)
    assert li.use_node(data) == timestamps

    # We are always representing timestamps as UTC
    assert li.use_node(data.toTimestamp()) == utc_dates


def test_mapeach_with_tags():
    data = [[2, 3, 4], [2, 3, 4], [2, 3, 4]]
    for i, row in enumerate(data):
        for j, elem in enumerate(row):
            row[j] = tag_store.add_tags(box.box(elem), {"tag": f"row{i}_col{j}"})
        data[i] = tag_store.add_tags(box.box(row), {"tag": f"row{i}"})
    tagged = tag_store.add_tags(box.box(data), {"tag": "top"})

    awl = arrow.to_arrow(tagged)
    node = weave.save(awl)

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("tag", types.String())

    result = arrow.ops.map_each(node, lambda row: row + 1)
    assert arrow.ArrowWeaveListType(types.List(types.Number())).assign_type(result.type)
    assert weave.use(result).to_pylist_notags() == [[3, 4, 5]] * 3

    # I believe there is a dispatch bug here. If we a regular list instead of awl
    # for node in this test, we'd get the single "top" tag here. The first dispatch
    # rule is "Prefer non-mapped". I think our intention is to return no list ops
    # there, but we don't exclude arrowweavelist ops. So we end up picking
    # the arrowweavelist ops, in the case where we have an arrow weave list input.
    # TODO fix
    assert weave.use(tag_getter_op(result)).to_pylist_notags() == [
        "row0",
        "row1",
        "row2",
    ]
    assert weave.use(tag_getter_op(result[0])) == "row0"
    assert weave.use(tag_getter_op(result[0][0])) == "row0_col0"


def test_unflatten_structs_in_flattened_table():
    flattened_table = pa.table(
        {
            "a➡️b➡️c": [1, 2, 3],
            "a➡️b➡️d": [4, 5, 6],
            "a➡️e": [7, 8, 9],
            "g": ["a", "b", "c"],
        }
    )
    result = arrow.ops.unflatten_structs_in_flattened_table(flattened_table)
    struct_result = pa.array(
        [
            {"a": {"b": {"c": 1, "d": 4}, "e": 7}, "g": "a"},
            {"a": {"b": {"c": 2, "d": 5}, "e": 8}, "g": "b"},
            {"a": {"b": {"c": 3, "d": 6}, "e": 9}, "g": "c"},
        ]
    )
    assert result == pa.table(
        {"a": struct_result.field("a"), "g": struct_result.field("g")}
    )


def test_map_with_index():
    ref = create_arrow_data(100)

    node = weave.get(ref).map(lambda row, index: index)
    assert weave.use(node).to_pylist_raw() == list(range(100))


def verify_pyarrow_array_type_is_valid_for_tag_array(pa_type: pa.DataType):
    if pa.types.is_string(pa_type):
        raise errors.WeaveInternalError(
            "Encountered invalid tag array, expected a DictionaryArray, got a StringArray"
        )
    elif pa.types.is_struct(pa_type):
        for field in pa_type:
            verify_pyarrow_array_type_is_valid_for_tag_array(field.type)
    elif pa.types.is_list(pa_type):
        verify_pyarrow_array_type_is_valid_for_tag_array(pa_type.value_type)


def test_verify_dictionary_encoding_of_strings():
    data = pa.array(["1", "2", "3"])

    # raises
    with pytest.raises(errors.WeaveInternalError):
        verify_pyarrow_array_type_is_valid_for_tag_array(data.type)

    data = pa.array([{"a": "1"}, {"a": "2"}, {"a": "3"}])

    # raises
    with pytest.raises(errors.WeaveInternalError):
        verify_pyarrow_array_type_is_valid_for_tag_array(data.type)

    converted = recursively_encode_pyarrow_strings_as_dictionaries(data)

    # does not raise
    verify_pyarrow_array_type_is_valid_for_tag_array(converted.type)

    data = pa.array([["1"], ["2"], ["3"]])
    # raises
    with pytest.raises(errors.WeaveInternalError):
        verify_pyarrow_array_type_is_valid_for_tag_array(data.type)

    converted = recursively_encode_pyarrow_strings_as_dictionaries(data)

    # does not raise
    verify_pyarrow_array_type_is_valid_for_tag_array(converted.type)


@pytest.mark.parametrize(
    "list_of_data, exp_res",
    [
        # Empty Case
        ([[]], []),
        # Base Case
        (
            [
                [{"a": 1, "b": 2}, {"a": 2, "b": 3}],
                [{"a": 4, "b": 5}, {"a": 6, "b": 7}],
            ],
            [{"a": 1, "b": 2}, {"a": 2, "b": 3}, {"a": 4, "b": 5}, {"a": 6, "b": 7}],
        ),
        # Deeply Mixed Case
        (
            [
                [{"a": {"c": 1, "d": 2}, "b": {"d": 3, "c": 4}}],
                [{"b": {"d": 5, "c": 6}, "a": {"c": 7, "d": 8}}],
            ],
            [
                {"a": {"c": 1, "d": 2}, "b": {"d": 3, "c": 4}},
                {"a": {"c": 7, "d": 8}, "b": {"c": 6, "d": 5}},
            ],
        ),
        # Empty Lists
        (
            [[{"a": [1]}], [{"a": []}], [{"a": None}]],
            [{"a": [1]}, {"a": []}, {"a": None}],
        ),
        # Mix of lists and dicts
        (
            [
                [{"a": [{"c": 1, "d": 2}, {"d": 3, "c": 4}]}],
                [{"a": [{"d": 5, "c": 6}, {"c": 7, "d": 8}]}],
            ],
            [
                {"a": [{"c": 1, "d": 2}, {"d": 3, "c": 4}]},
                {"a": [{"d": 5, "c": 6}, {"c": 7, "d": 8}]},
            ],
        ),
        # Nested Mixed Types
        (
            [
                [{"a": 1, "b": 2, "c": 4}, {"a": "a", "b": 3, "c": 5}],
                [{"a": 4, "b": 5, "c": "c"}, {"a": 6, "b": 7}],
            ],
            [
                {"a": 1, "b": 2, "c": 4},
                {"a": "a", "b": 3, "c": 5},
                {"a": 4, "b": 5, "c": "c"},
                {"a": 6, "b": 7, "c": None},
            ],
        ),
        # Nested Mixed Types (flat w/ varying depth)
        (
            [
                [{"a": 1}],
                [{"a": [1]}],
                [{"a": "a"}],
                [{"a": ["a"]}],
                [{"a": {"b": 1}}],
                [{"a": [{"b": 1}]}],
            ],
            [
                {"a": 1},
                {"a": [1]},
                {"a": "a"},
                {"a": ["a"]},
                {"a": {"b": 1}},
                {"a": [{"b": 1}]},
            ],
        ),
    ],
)
def test_arrow_concat_mixed(list_of_data, exp_res):
    assert (
        weave.use(
            ops.make_list(
                **{
                    f"{ndx}": weave.save(arrow.to_arrow(data))
                    for ndx, data in enumerate(list_of_data)
                }
            ).concat()
        ).to_pylist_raw()
        == exp_res
    )


def test_complex_concat_union():
    l_0_0 = weave.save(arrow.to_arrow([{"a": [1]}]))
    l_0_1 = weave.save(arrow.to_arrow([{"a": []}]))
    l_0_2 = weave.save(arrow.to_arrow([{"a": None}]))
    l_0 = ops.make_list(a=l_0_0, b=l_0_1, c=l_0_2).concat()["a"]

    # This is intentionally empty!
    l_1_0 = weave.save(arrow.to_arrow([{"a": []}]))
    l_1_1 = weave.save(arrow.to_arrow([{"a": []}]))
    l_1_2 = weave.save(arrow.to_arrow([{"a": None}]))
    l_1 = ops.make_list(a=l_1_0, b=l_1_1, c=l_1_2).concat()["a"]

    l = ops.make_list(a=l_0, b=l_1).concat()

    assert weave.use(l).to_pylist_raw() == [[1], [], None, [], [], None]


def test_abs():
    data = [-10, -2.2, 5, None, 3.3]
    arrow_node = weave.save(arrow.to_arrow(data))
    assert weave.use(arrow_node.abs()).to_pylist_raw() == [10, 2.2, 5, None, 3.3]


def test_argmax():
    data = [10, 20, None, 30, 40, None, 30, 20, None, 10]
    arrow_node = weave.save(arrow.to_arrow(data))
    assert weave.use(arrow_node.argmax()) == 4


def test_argmin():
    data = [10, 20, None, 30, 40, None, 0, 20, None, 10]
    arrow_node = weave.save(arrow.to_arrow(data))
    assert weave.use(arrow_node.argmin()) == 6


def test_stddev():
    data = [10, 20, None, 30, 40, None, 0, 20, None, 10]
    arrow_node = weave.save(arrow.to_arrow(data))
    res = round(weave.use(arrow_node.stddev()), 3)
    assert res == 12.454


def test_join_all_struct_val():
    from weave.legacy.weave import ops_arrow

    t1 = arrow.to_arrow([{"a": 5, "b": {"c": 6}}])
    t2 = arrow.to_arrow([{"a": 9, "b": {"c": 10}}, {"a": 5, "b": {"c": 11}}])

    tables = weave.save([t1, t2])
    joined = tables.joinAll(lambda row: row["a"], True)

    res = weave.use(joined).to_pylist_raw()
    # TODO: not correct, not because of join, because artifact saving is broken.
    assert res == [
        {"_tag": {"joinObj": 5}, "_value": {"a": [5, 5], "b": [{"c": 6}, {"c": 11}]}},
        {"_tag": {"joinObj": 9}, "_value": {"a": [None, 9], "b": [None, {"c": 10}]}},
    ]


def test_join_all_on_list():
    t1 = arrow.to_arrow([{"a": [5], "b": {"c": 6}}])
    t2 = arrow.to_arrow([{"a": [9], "b": {"c": 10}}, {"a": [5], "b": {"c": 11}}])

    tables = weave.save([t1, t2])
    joined = tables.joinAll(lambda row: row["a"], True)
    res = weave.use(joined).to_pylist_raw()
    # TODO: not correct, not because of join, because artifact saving is broken.
    assert res == [
        {
            "_tag": {"joinObj": [5]},
            "_value": {"a": [[5], [5]], "b": [{"c": 6}, {"c": 11}]},
        },
        {
            "_tag": {"joinObj": [9]},
            "_value": {"a": [None, [9]], "b": [None, {"c": 10}]},
        },
    ]


def test_to_arrow_union_list():
    val = [{"a": 5.0}, {"a": [1.0]}]
    arrow_val = arrow.to_arrow([{"a": 5.0}, {"a": [1.0]}])
    assert arrow_val.to_pylist_raw() == val


def test_concat_empty_arrays():
    val = arrow.to_arrow([])
    val2 = arrow.to_arrow([{"a": 5}])
    assert val.concat(val).to_pylist_raw() == []
    assert val.concat(val2).to_pylist_raw() == val2.to_pylist_raw()
    assert val2.concat(val).to_pylist_raw() == val2.to_pylist_raw()
    assert val2.concat(val2).to_pylist_raw() == [{"a": 5}, {"a": 5}]


_loading_builtins_token = context_state.set_loading_built_ins()


def _test_arrow_do_body(a: int, b: int, c: list[int]) -> int:
    return a * b + a ** c[0]


@weave.op()
def _test_arrow_do_op(a: int, b: int, c: list[int]) -> int:
    if isinstance(a, graph.Node):
        raise errors.WeavifyError("weavifying")
    return _test_arrow_do_body(a, b, c)


context_state.clear_loading_built_ins(_loading_builtins_token)


def test_concat_nulls():
    datal = weave.save(
        arrow.to_arrow([{"prompt": None}, {"prompt": None}, {"prompt": None}])
    )
    datar = weave.save(
        arrow.to_arrow(
            [
                {"prompt": ["a"]},
                {"prompt": ["d"]},
                {"prompt": ["e"]},
            ]
        )
    )

    list_nodes = make_list(l1=datal, l2=datar)
    concatenated = arrow.ops.concat(list_nodes)

    assert weave.use(concatenated.to_py()) == [
        {"prompt": None},
        {"prompt": None},
        {"prompt": None},
        {"prompt": ["a"]},
        {"prompt": ["d"]},
        {"prompt": ["e"]},
    ]


def test_automap_more_than_one():
    data = [1, 2, -5, -100]
    arrow_data = weave.save(arrow.to_arrow(data))

    assert weave.use(
        arrow_data.map(lambda x: _test_arrow_do_op(x, x, [4]))
    ).to_pylist_raw() == [_test_arrow_do_body(x, x, [4]) for x in data]

    assert weave.use(
        arrow_data.map(lambda x: _test_arrow_do_op(x + 1, x, [4]))
    ).to_pylist_raw() == [_test_arrow_do_body(x + 1, x, [4]) for x in data]

    assert weave.use(
        arrow_data.map(lambda x: _test_arrow_do_op(x + 1, x, ops.make_list(a=x * 2)))
    ).to_pylist_raw() == [_test_arrow_do_body(x + 1, x, [x * 2]) for x in data]

    assert weave.use(
        arrow_data.map(lambda x: _test_arrow_do_op(-1, x, ops.make_list(a=x * 2)))
    ).to_pylist_raw() == [_test_arrow_do_body(-1, x, [x * 2]) for x in data]

    assert weave.use(
        arrow_data.map(lambda x: _test_arrow_do_op(-1, -9, ops.make_list(a=x * 2)))
    ).to_pylist_raw() == [_test_arrow_do_body(-1, -9, [x * 2]) for x in data]


def test_serialization_of_boxed_nones():
    data = [{"a": 1, "b": box.box(None)}, {"a": 2, "b": box.box(None)}]
    arrow_data = weave.save(arrow.to_arrow(data))
    assert weave.use(arrow_data).to_pylist_raw() == data


def test_join_tag_support():
    # tagged<list<tagged<awl<tagged<dict<id: tagged<number>, col: tagged<string>>>>>>>
    cst = weave_internal.const
    tag = ttu.op_add_tag

    def cst_list(data):
        return ops.make_list(**{str(i): l for i, l in enumerate(data)})

    def cst_dict(data):
        return ops.dict_(**data)

    lists_rows = []
    for l_i in range(3):
        list_rows = []
        for r_i in range(4):
            row_key = f"{l_i}_{r_i}"
            # Purposely having id duplicated for purposes of joining
            list_rows.append(
                tag(
                    cst_dict(
                        {
                            "id": tag(
                                cst(f"id_val_{r_i}"), {"id_tag": f"id_{row_key}"}
                            ),
                            "col": tag(
                                cst(f"col_val_{row_key}"), {"col_tag": f"col_{row_key}"}
                            ),
                        }
                    ),
                    {"row_tag": f"row_{row_key}"},
                )
            )
        raw_list = cst_list(list_rows)
        raw_awl = arrow.ops.list_to_arrow(raw_list)
        tagged_awl = tag(raw_awl, {"awl_tag": f"awl_{l_i}"})
        lists_rows.append(tagged_awl)
    raw_lists = cst_list(lists_rows)

    get_id = ttu.make_get_tag("id_tag")
    get_col = ttu.make_get_tag("col_tag")
    get_row = ttu.make_get_tag("row_tag")
    get_awl = ttu.make_get_tag("awl_tag")

    second_list = weave.use(raw_lists[1])
    assert second_list.to_pylist_raw() == [
        {
            "_tag": {"_ct_row_tag": "row_1_0"},
            "_value": {
                "id": {"_tag": {"_ct_id_tag": "id_1_0"}, "_value": "id_val_0"},
                "col": {"_tag": {"_ct_col_tag": "col_1_0"}, "_value": "col_val_1_0"},
            },
        },
        {
            "_tag": {"_ct_row_tag": "row_1_1"},
            "_value": {
                "id": {"_tag": {"_ct_id_tag": "id_1_1"}, "_value": "id_val_1"},
                "col": {"_tag": {"_ct_col_tag": "col_1_1"}, "_value": "col_val_1_1"},
            },
        },
        {
            "_tag": {"_ct_row_tag": "row_1_2"},
            "_value": {
                "id": {"_tag": {"_ct_id_tag": "id_1_2"}, "_value": "id_val_2"},
                "col": {"_tag": {"_ct_col_tag": "col_1_2"}, "_value": "col_val_1_2"},
            },
        },
        {
            "_tag": {"_ct_row_tag": "row_1_3"},
            "_value": {
                "id": {"_tag": {"_ct_id_tag": "id_1_3"}, "_value": "id_val_3"},
                "col": {"_tag": {"_ct_col_tag": "col_1_3"}, "_value": "col_val_1_3"},
            },
        },
    ]

    joined = raw_lists.joinAll(lambda row: row["id"], True)
    joined_row = joined[1]

    assert weave.use(joined_row) == {
        "id": ["id_val_1", "id_val_1", "id_val_1"],
        "col": ["col_val_0_1", "col_val_1_1", "col_val_2_1"],
    }

    joined_row_item = joined_row["col"][1]

    assert (weave.use(joined_row_item)) == "col_val_1_1"
    assert (weave.use(get_col(joined_row_item))) == "col_1_1"
    assert (weave.use(get_row(joined_row_item))) == "row_1_1"
    assert (weave.use(get_awl(joined_row_item))) == "awl_1"

    joined_row_obj = joined_row_item.joinObj()

    assert (weave.use(joined_row_obj)) == "id_val_1"


def test_arrow_handling_of_empty_structs():
    data = [{"a": 5, "b": {}}, {"a": 6, "b": {}}, {"a": 7, "b": None}]
    arrow_data = weave.save(arrow.to_arrow(data))
    assert weave.use(arrow_data.map(lambda x: x["b"])).to_pylist_raw() == [{}, {}, None]
    assert weave.use(arrow_data.map(lambda x: x["a"])).to_pylist_raw() == [5, 6, 7]


def test_count_on_group():
    col = weave.save(arrow.to_arrow([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]))
    col = col.groupby(lambda x: x % 2)
    col = col.map(lambda x: x.count())
    assert weave.use(col).to_pylist_raw() == [
        {"_tag": {"groupKey": 1}, "_value": 5},
        {"_tag": {"groupKey": 0}, "_value": 5},
    ]


def test_limit_on_group():
    col = weave.save(arrow.to_arrow([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]))
    col = col.groupby(lambda x: x % 2)
    col = col.map(lambda x: x.limit(3))
    assert weave.use(col).to_pylist_raw() == [[1, 3, 5], [2, 4, 6]]


def test_offset_on_group():
    col = weave.save(arrow.to_arrow([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]))
    col = col.groupby(lambda x: x % 2)
    col = col.map(lambda x: x.offset(2))
    assert weave.use(col).to_pylist_raw() == [[5, 7, 9], [6, 8, 10]]


def test_map_column():
    arr = arrow.to_arrow(
        [{"a": {"c": 6, "d": 7}, "b": 9}, {"a": {"c": 14, "d": 7}, "b": 5}]
    )

    def _map(awl, path):
        if isinstance(awl.object_type, types.TypedDict):
            # Keep the first element of the dict
            key0, type0 = next(iter(awl.object_type.property_types.items()))
            return arrow.ArrowWeaveList(
                awl._arrow_data.field(key0),
                type0,
                awl._artifact,
            )

    res = arr.map_column(_map)
    # Since this is post-order traversal of the type tree, we should
    # only see the innermost type
    assert res.object_type == types.Int()
    assert res.to_pylist_raw() == [6, 14]


def test_encode_decode_list_of_dictionary_encoded_strings():
    data = [["a", "b", "c"], ["b", "c"], ["a", "b", None], None, ["d"], []]
    awl = arrow.to_arrow(
        data,
        wb_type=types.List(
            types.UnionType(
                types.List(types.UnionType(types.String(), types.NoneType())),
                types.NoneType(),
            )
        ),
    )
    awl._arrow_data = recursively_encode_pyarrow_strings_as_dictionaries(
        awl._arrow_data
    )
    result = awl.to_pylist_raw()
    assert result == data


def test_pushdown_of_gql_tags_on_awls(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: test_wb.workspace_response())
    project_node = ops.project("stacey", "mendeleev")
    project = weave.use(project_node)
    data = box.box([1, 2, 3])
    tag_store.add_tags(data, {"project": project})
    awl = weave.save(arrow.to_arrow(data))

    list_node = arrow.ops.arrow_list_(**{"a": awl, "b": awl})
    concatted = arrow.ops.concat(list_node)

    for i in range(6):
        cell = arrow.ops.index(concatted, i)
        ptag = project_ops.get_project_tag(cell)
        assert weave.use(ptag) == project

    count = concatted.count()
    assert weave.use(count) == 6


def test_groupby_concat():
    # query at issue
    #   some-table().rows().concat()
    #   .groupby(some_function)
    #   .dropna().concat()'  <- note second concat here

    data = [{"time": 1, "model_type": "a"}, {"time": 2, "model_type": "b"}]
    awl_node = weave.save(arrow.to_arrow(data))

    # concat
    list_node = ops.make_list(**{str(i): awl_node for i in range(4)})
    concatted = arrow.ops.concat(list_node)

    # groupby
    grouped = concatted.groupby(lambda x: ops.dict_(**{"time": x["time"]}))

    # dropna
    dropped = grouped.dropna()

    # now concat all the groups together
    concatted_2 = arrow.ops.concat(dropped)
    selected = concatted_2.map(
        lambda row: ops.dict_(time=row.groupkey()["time"], model_type=row["model_type"])
    )

    result = weave.use(selected).to_pylist_notags()
    assert result == ([data[0]] * 4) + ([data[1]] * 4)


def test_conversion_of_domain_types_to_awl_values(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: test_wb.workspace_response())
    project_node = ops.project("stacey", "mendeleev")
    project = weave.use(project_node)
    data = box.box([project] * 3)
    awl = weave.save(arrow.to_arrow(data))

    list_node = arrow.ops.arrow_list_(**{"a": awl, "b": awl})
    assert [[item for item in l] for l in weave.use(list_node)] == [[project] * 2] * 3


def test_non_zero_offset():
    non_zero_offset_array = pa.array([[0, 1], [2, 3], [4, 5]]).slice(1)
    awl = arrow.ArrowWeaveList(non_zero_offset_array)

    # First assertion validates that the test setup is correct:
    assert awl._arrow_data.offsets[0] != 0

    # We just want to validate that this function completes - no assertion
    # needed
    awl.map_column(lambda x, y: None)


def test_object_types_nullable():
    data_node = arrow.to_weave_arrow(
        [
            {"a": 5, "point": Point2(256, 256)},
            {"a": 6, "point": None},
        ]
    )
    assert weave.use(data_node[0]["point"].get_x()) == 256

    assert weave.use(
        data_node.map(lambda row: row["point"].get_x())
    ).to_pylist_raw() == [
        256,
        None,
    ]


def test_save_nested_custom_objs():
    t1 = arrow.to_arrow([{"a": 5}])
    t2 = arrow.to_arrow([{"a": 9}])

    tables = weave.save([t1, t2])
    assert weave.use(tables[0]).to_pylist_raw() == [{"a": 5}]
    assert weave.use(tables[1]).to_pylist_raw() == [{"a": 9}]


def test_to_compare_safe():
    l = [[], "a", 5]
    a = arrow.to_arrow(l)
    safe = arrow.to_compare_safe(a)
    assert safe.to_pylist_notags() == ["__t_13-__list_-", "__t_13-a", "__t_9-5"]


def test_empty_dict():
    data = [{}]
    awl = arrow.to_arrow(data)
    ref = storage.save(awl)
    awl2 = storage.get(str(ref))
    assert len(awl2) == 1
    assert awl._arrow_data == awl2._arrow_data


def test_arrow_op_decorator_handles_optional_tagged_type():
    obj = [None, 1, 2, None]
    obj[1] = box.box(obj[1])
    obj[2] = box.box(obj[2])
    tag_store.add_tags(obj[1], {"mytag": "a"})
    tag_store.add_tags(obj[2], {"mytag": "b"})
    awl = arrow.to_arrow(obj)
    assert awl.object_type == types.optional(
        tagged_value_type.TaggedValueType(
            types.TypedDict({"mytag": types.String()}), types.Int()
        )
    )

    node = weave.save(awl)
    toTimestamp = node.toTimestamp()

    expected = [
        None,
        datetime.datetime(1970, 1, 1, 0, 0, 0, 1000, tzinfo=datetime.timezone.utc),
        datetime.datetime(1970, 1, 1, 0, 0, 0, 2000, tzinfo=datetime.timezone.utc),
        None,
    ]

    expected[1] = box.box(expected[1])
    expected[2] = box.box(expected[2])

    tag_store.add_tags(expected[1], {"mytag": "a"})
    tag_store.add_tags(expected[2], {"mytag": "b"})

    expected = arrow.to_arrow(expected)
    actual = weave.use(toTimestamp)

    assert actual.to_pylist_raw() == expected.to_pylist_raw()


def test_flatten_handles_tagged_lists():
    data = [[1], [2, 3], [4, 5, 6]]
    for i in range(len(data)):
        data[i] = box.box(data[i])
        tag_store.add_tags(data[i], {"outer": "outer"})
        for j in range(len(data[i])):
            data[i][j] = box.box(data[i][j])
            tag_store.add_tags(data[i][j], {"inner": "inner"})

    awl = arrow.to_arrow(data)
    node = weave.save(awl)
    flattened = node.flatten()

    assert flattened.type == arrow.ArrowWeaveListType(
        tagged_value_type.TaggedValueType(
            types.TypedDict({"outer": types.String(), "inner": types.String()}),
            types.Int(),
        )
    )

    expected = [1, 2, 3, 4, 5, 6]
    assert weave.use(flattened).to_pylist_notags() == expected
    assert weave.use(flattened).to_pylist_raw() == [
        {
            "_tag": {
                "outer": "outer",
                "inner": "inner",
            },
            "_value": i,
        }
        for i in expected
    ]


def test_keys_ops():
    awl = arrow.to_arrow([{"a": 1}, {"a": 1, "b": 2, "c": 2}, {"c": 3}])
    node = weave.save(awl)
    keys_node = node.keys()
    # Unfortunately, we lose specific info about key presence in AWL.
    assert weave.use(keys_node).to_pylist_raw() == [
        ["a", "b", "c"],
        ["a", "b", "c"],
        ["a", "b", "c"],
    ]

    all_keys_node = keys_node.flatten().unique()

    assert weave.use(all_keys_node).to_pylist_raw() == ["a", "b", "c"]


def test_repeat_0():
    data = {"a": 1}
    repeated = constructors.repeat(data, 0)
    assert len(repeated) == 0
    assert repeated.type == pa.struct({"a": pa.int64()})



import typing


class RunKey(typing.Protocol):
    @property
    def id(self) -> str:
        ...

    @property
    def trace_id(self) -> str:
        ...


class Run(typing.Protocol):
    @property
    def id(self) -> str:
        ...

    @property
    def trace_id(self) -> str:
        ...

    @property
    def ui_url(self) -> str:
        ...
